#ifndef	_REENTRANT
#define	_REENTRANT
#endif

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <netdb.h>
#include <values.h>
#include <time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/stream.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/dlpi.h>

#include "ppp.h"
#include "dp.h"
#include "dpio.h"
#include "dpd.h"
#include "dpconf.h"


#define	MAXLN	512
#define	MAXTOK	256

static access_t		*AccessList;

#define EQ(a, b)	(strcasecmp((a), (b)) == 0)
#define EQN(a, b, n)	(strncasecmp((a), (b), (n)) == 0)

#define nexttok(pat, lt)	strtok_r((char *)NULL, pat, lt)
static char		SPACES[]  = " \t";
static char		SLASH[] = "/";
static char		DOTSLASH[] = "./";
static char		COMMA[]  = ",";
static char		RBRACK[]  = "]";
static char		SPECIAL[] = "(){}|&!#";

typedef struct mapping {
    long	Number;
    char	*Name;
} mapping_t;

#define	QUAL_ALLOW	1
#define	QUAL_DISALLOW	2
#define	QUAL_CONTINUOUS	3

static mapping_t cond_tab[] = {
    { COND_PROTO, "protos" },
    { COND_ADDR, "addrs" },
    { COND_TIME, "times" },
    { 0, (char *)0 }
};

static mapping_t qual_tab[] = {
    { COND_SRC, "src" },
    { COND_DST, "dst" },
    { 0, (char *)0 }
};

static mapping_t allow_tab[] = {
    { 0, "allow" },
    { 1, "disallow" },
    { 0, (char *)0 }
};

#define	TIMES_CALL	1
#define	TIMES_CONNECT	2

static mapping_t time_tab[] = {
    { TIMES_CALL,	"call" },
    { TIMES_CONNECT,	"connect" },
    { 0,		(char *)0 }
};


static mapping_t op_tab[] = {
    { OP_AND,	"&&"	},
    { OP_AND,	"and"	},
    { OP_OR,	"||"	},
    { OP_OR,	"or"	},
    { OP_NOT,	"!"	},
    { OP_NOT,	"not"	},
    { OP_LPAR,	"("	},
    { OP_RPAR,	")"	},
    { OP_LCBK,	"{"	},
    { OP_RCBK,	"}"	},
    { OP_TRUE,	"true"	},
    { OP_FALSE,	"false"	},
    { 0,	(char *)0 }
};

#define	TIMEOUT_BOTH	0
#define	TIMEOUT_DIAL	1
#define	TIMEOUT_ANSWER	2

static mapping_t timeout_tab[] = {
    { TIMEOUT_DIAL,	"dial" },
    { TIMEOUT_ANSWER,	"answer" },
    { 0,		(char *)0 }
};

static mapping_t to_which_tab[] = {
    { TO_FAILCALL,  "failedcall"},
    { TO_WAITING,   "callwait"	},
    { TO_ACTIVE,    "inactivity"},
    { TO_ACTIVE_C,  "last_close"},
    { TO_ACTIVE_U,  "non_stream"},
    { 0,	    (char *)0	}
};

#ifdef	__STDC__
static access_t *		alloc_access_file(char *);
void				free_access_files();
static void			free_access_file(access_t *);
static condition_list_t *	alloc_condition_list();
static condition_t *		alloc_condition(int);
static void			free_condition(condition_t *);
static void			free_pattern(pattern_t *);
static range_t *		alloc_range();
static void			free_range(range_t *);
static timeout_t *		alloc_timeout();
static path_t *			alloc_path(int);
static path_t *			new_path(int, int, long *);
static void			free_path(path_t *);
static pkt_info_t *		alloc_pkt_info();
static void			free_pkt_info(pkt_info_t *);
static void			free_timeout(timeout_t *);
static void			add_condition_list(timeout_t *,
						   condition_list_t *);
static int			any_timeos(dp_timeo_t *, int);
static void			dp_settimeos(dp_timeo_t *, int);
static void			dp_maptimeos(dp_timeo_t *, int, int);
static void			dp_addtimeos(dp_timeo_t *, dp_timeo_t *);
static int			commentline(char *);
static char *			agetln(char *, int);
static void			askipln();
static void			aputbacktok(char *);
static char *			agettok();
static int			parse_access(access_t *, int);
int				readaccess(char *, char *, REMOTE *);
static char **			tok_vector(char *, int, int *);
static void			free_condition(condition_t *);
static void			rm_tok_from_vec(char **, int *, int);
static int			parse_access_v2(access_t *);
static int			parse_seconds_sub(char *, char **);
static int			parse_seconds(char *);
static void			parse_timeout(dp_timeo_t *, int);
static mapping_t *		getptabbyname(char *, mapping_t *);
static mapping_t *		getptabbynum(long, mapping_t *);
static long			mapname2num(char *, mapping_t *);
static char *			mapnum2name(long, mapping_t *);
static int			match_rangelist(long, range_t *);
char *				ppath_str(path_t *);
char *				apath_str(path_t *);
char *				tpath_str(path_t *);
char *				time_str(long, int);
char *				path_str(path_t *);
char *				appath_str(path_t *, path_t *);
char *				addr_proto_str(pkt_info_t *);
static int			match_pat(path_t *, condition_t *);
static char *			lcase(char *);
static void			set_range(range_t *, int, long);
static void			map_proto_element(range_t *, int, char *, int,
						  condition_t *);
static void			map_addr_element(range_t *, int, char *, int,
						 condition_t *);
static void			map_time_element(range_t *, int, char *, int,
						 condition_t *);
static char *			agetln(char *, int);
static void			expand_time_list(range_t *, int);
static range_t *		parse_range(char *, void (*)(), int,
					    condition_t *);
static char *			next_range(char **);
static char *			first_range(char *, char **);
static range_t *		parse_range_list(char *, void (*)(), int,
						 condition_t *);
static condition_t *		parse_proto_path(char *, int);
static pattern_t *		inet_path(char *, int);
static condition_t *		addr2path(char *, int);
static condition_t *		parse_addr_path(char *, int);
static condition_t *		parse_condition_expr(int, int);
static condition_t *		parse_condition_cbrk(int, int);
static condition_t *		parse_time_path(char *);
static char *			strip_qual(char **);
static int			prefix_colon(char **, mapping_t *);
static condition_t *		parse_path(char *, int);
static void			parse_timeouts(timeout_t **);
static int			ispositivenumber(char *);
void				dlsap2ph(phdr_t *, ulong);
static path_t *			packet2proto(dl_unitdata_ind_t *, char *, int);
static path_t *			packet2addr(dl_unitdata_ind_t *, char *, int);
static path_t *			time_path(time_t);
static path_t *			current_time_path();
static pkt_info_t *		pkt_info(dl_unitdata_ind_t *, char *pkt);
int				authorized(access_t *, dl_unitdata_ind_t *,
					   int, char *, int);
static int			match_cond(condition_t *, pkt_info_t *);
static void			timeout_rules(dp_timeo_t *, timeout_t *,
					      pkt_info_t *);
void				comp_timeos(dp_timeo_t *, access_t *, int,
					    dl_unitdata_ind_t *, char *);
static int			set_timeouts(REMOTE *, int,
					     dl_unitdata_ind_t *, char *);
int				set_dtimeouts(REMOTE *, dl_unitdata_ind_t *,
					      char *);
int				set_atimeouts(REMOTE *);
static condition_t *		parse_condition_orlist(int);
static condition_t *		parse_condition_list(int, int);
static void			parse_timeout_v1(dp_timeo_t *, int);
static condition_t *		parse_time_list(char *);
static void			add_time_cond(condition_t **, condition_t *);
static int			parse_access_v1(access_t *);
#else
static access_t *		alloc_access_file(/* char * */);
void				free_access_files(/*  */);
static void			free_access_file(/* access_t * */);
static condition_list_t *	alloc_condition_list(/*  */);
static condition_t *		alloc_condition(/* int */);
static void			free_condition(/* condition_t * */);
static void			free_pattern(/* pattern_t * */);
static range_t *		alloc_range(/*  */);
static void			free_range(/* range_t * */);
static timeout_t *		alloc_timeout(/*  */);
static path_t *			alloc_path(/* int */);
static path_t *			new_path(/* int, int, long * */);
static void			free_path(/* path_t * */);
static pkt_info_t *		alloc_pkt_info(/*  */);
static void			free_pkt_info(/* pkt_info_t * */);
static void			free_timeout(/* timeout_t * */);
static void			add_condition_list(/* timeout_t *,
						   condition_list_t * */);
static int			any_timeos(/* dp_timeo_t *, int */);
static void			dp_settimeos(/* dp_timeo_t *, int */);
static void			dp_maptimeos(/* dp_timeo_t *, int, int */);
static void			dp_addtimeos(/* dp_timeo_t *, dp_timeo_t * */);
static int			commentline(/* char * */);
static char *			agetln(/* char *, int */);
static void			askipln(/*  */);
static void			aputbacktok(/* char * */);
static char *			agettok(/*  */);
static int			parse_access(/* access_t *, int */);
int				readaccess(/* char *, char *, REMOTE * */);
static char **			tok_vector(/* char *, int, int * */);
static void			free_condition(/* condition_t * */);
static void			rm_tok_from_vec(/* char **, int *, int */);
static int			parse_access_v2(/* access_t * */);
static int			parse_seconds_sub(/* char *, char ** */);
static int			parse_seconds(/* char * */);
static void			parse_timeout(/* dp_timeo_t *, int */);
static mapping_t *		getptabbyname(/* char *, mapping_t * */);
static mapping_t *		getptabbynum(/* long, mapping_t * */);
static long			mapname2num(/* char *, mapping_t * */);
static char *			mapnum2name(/* long, mapping_t * */);
static int			match_rangelist(/* long, range_t * */);
char *				ppath_str(/* path_t * */);
char *				apath_str(/* path_t * */);
char *				tpath_str(/* path_t * */);
char *				time_str(/* long, int */);
char *				path_str(/* path_t * */);
char *				appath_str(/* path_t *, path_t * */);
char *				addr_proto_str(/* pkt_info_t * */);
static int			match_pat(/* path_t *, condition_t * */);
static char *			lcase(/* char * */);
static void			set_range(/* range_t *, int, long */);
static void			map_proto_element(/* range_t *, int, char *, int,
						  condition_t * */);
static void			map_addr_element(/* range_t *, int, char *, int,
						 condition_t * */);
static void			map_time_element(/* range_t *, int, char *, int,
						 condition_t * */);
static char *			agetln(/* char *, int */);
static void			expand_time_list(/* range_t *, int */);
static range_t *		parse_range(/* char *, void (*)(), int,
					    condition_t * */);
static char *			next_range(/* char ** */);
static char *			first_range(/* char *, char ** */);
static range_t *		parse_range_list(/* char *, void (*)(), int,
						 condition_t * */);
static condition_t *		parse_proto_path(/* char *, int */);
static pattern_t *		inet_path(/* char *, int */);
static condition_t *		addr2path(/* char *, int */);
static condition_t *		parse_addr_path(/* char *, int */);
static condition_t *		parse_condition_expr(/* int, int */);
static condition_t *		parse_condition_cbrk(/* int, int */);
static condition_t *		parse_time_path(/* char * */);
static char *			strip_qual(/* char ** */);
static int			prefix_colon(/* char **, mapping_t * */);
static condition_t *		parse_path(/* char *, int */);
static void			parse_timeouts(/* timeout_t ** */);
static int			ispositivenumber(/* char * */);
void				dlsap2ph(/* phdr_t *, ulong */);
static path_t *			packet2proto(/* dl_unitdata_ind_t *, char *, int */);
static path_t *			packet2addr(/* dl_unitdata_ind_t *, char *, int */);
static path_t *			time_path(/* time_t */);
static path_t *			current_time_path(/*  */);
static pkt_info_t *		pkt_info(/* dl_unitdata_ind_t *, char *pkt */);
int				authorized(/* access_t *, dl_unitdata_ind_t *,
					   int, char *, int */);
static int			match_cond(/* condition_t *, pkt_info_t * */);
static void			timeout_rules(/* dp_timeo_t *, timeout_t *,
					      pkt_info_t * */);
void				comp_timeos(/* dp_timeo_t *, access_t *, int,
					    dl_unitdata_ind_t *, char * */);
static int			set_timeouts(/* REMOTE *, int,
					     dl_unitdata_ind_t *, char * */);
int				set_dtimeouts(/* REMOTE *, dl_unitdata_ind_t *,
					      char * */);
int				set_atimeouts(/* REMOTE * */);
static condition_t *		parse_condition_orlist(/* int */);
static condition_t *		parse_condition_list(/* int, int */);
static void			parse_timeout_v1(/* dp_timeo_t *, int */);
static condition_t *		parse_time_list(/* char * */);
static void			add_time_cond(/* condition_t **, condition_t * */);
static int			parse_access_v1(/* access_t * */);
#endif

extern int errorcount,
	   linenum;


static access_t *
alloc_access_file(path)
char *path;
{
    access_t *ap;
    ap = (access_t *)calloc(1, sizeof(*ap));
    ap->Path = path;
    ap->Next = AccessList;
    AccessList = ap;
    return ap;
}

void
free_access_files()
{
    access_t *ap, *apn;

    for (ap = AccessList ; ap ; ap = apn) {
	apn = ap->Next;
	free_access_file(ap);
    }
    AccessList = (access_t *)0;
}

static void
free_access_file(ap)
access_t *ap;
{
    free_condition(ap->Protos);
    free_condition(ap->Addrs);
    free_condition(ap->CallTimes);
    free_condition(ap->ConnectTimes);
    free_timeout(ap->DTimeouts);
    free_timeout(ap->ATimeouts);
    free_timeout(ap->Timeouts);
    (void)free(ap->Path);
    (void)free((char *)ap);
}

static condition_list_t *
alloc_condition_list()
{
    condition_list_t *cl;
    cl = (condition_list_t *)calloc(1, sizeof(condition_list_t));
    dp_settimeos(&cl->Timeouts, DPT_UNSPEC_TIMEOUT);
    return cl;
}

static condition_t *
alloc_condition(type)
int type;
{
    condition_t *c;
    c = (condition_t *)calloc(1, sizeof(condition_t));
    c->Type = type;
    return c;
}

static void
free_condition(c)
condition_t *c;
{
    if (!c)
	return;
    free_condition(c->Left);
    free_condition(c->Right);
    free_pattern(c->Pattern);
    free((char *)c);
}

static pattern_t *
alloc_pattern()
{
    return (pattern_t *)calloc(1, sizeof(pattern_t));
}

static void
free_pattern(p)
pattern_t *p;
{
    if (!p)
	return;
    free_pattern(p->NextLevel);
    free_range(p->RangeList);
    free((char *)p);
}

static range_t *
alloc_range()
{
    return (range_t *)calloc(1, sizeof(range_t));
}

static void
free_range(r)
range_t *r;
{
    if (!r)
	return;
    free_range(r->Next);
    (void)free((char *)r);
}

static timeout_t *
alloc_timeout()
{
    timeout_t *to;
    to = (timeout_t *)calloc(1, sizeof(timeout_t));
    to->DefaultCondition = alloc_condition_list();
    to->LastConditionList = &to->ConditionList;
    return to;
}

static path_t *
alloc_path(n)
int n;
{
    path_t *p = malloc(sizeof(*p));
    p->NLevels = n;
    p->Levels = malloc(n * sizeof(*p->Levels));
    return p;
}

static path_t *
new_path(type, n, plevs)
int type,
    n;
long *plevs;
{
    path_t *p = alloc_path(n);
    p->Type = type;
    (void)memcpy((char *)p->Levels, (char *)plevs, sizeof(*p->Levels)*n);
    return p;
}

static void
free_path(p)
path_t *p;
{
    (void)free(p->Levels);
    (void)free(p);
}

static pkt_info_t *
alloc_pkt_info()
{
    return (pkt_info_t *)calloc(1, sizeof(pkt_info_t));
}

static void
free_pkt_info(pi)
pkt_info_t *pi;
{
    if (pi->DstProto)
	free_path(pi->DstProto);
    if (pi->SrcProto)
	free_path(pi->SrcProto);
    if (pi->DstAddr)
	free_path(pi->DstAddr);
    if (pi->SrcAddr)
	free_path(pi->SrcAddr);
    if (pi->CurrentTime)
	free_path(pi->CurrentTime);
    free((char *)pi);
}

static void
free_timeout(to)
timeout_t *to;
{
    free((char *)to);
}

static void
add_condition_list(to, cl)
timeout_t *to;
condition_list_t *cl;
{
    cl->Next = *to->LastConditionList;
    *to->LastConditionList = cl;
    while (cl->Next)
	cl = cl->Next;
    to->LastConditionList = &cl->Next;
}

static void
add_default_condition(to)
timeout_t *to;
{
    if (!to)
	return;
    add_condition_list(to, to->DefaultCondition);
}

static char *access_file;
static FILE *access_fp;

static int
any_timeos(dt, to)
dp_timeo_t *dt;
{
    return (to == dt->dt_active)
        || (to == dt->dt_active_c)
        || (to == dt->dt_active_u)
        || (to == dt->dt_waiting)
        || (to == dt->dt_failcall);
}

static void
dp_settimeos(dt, to)
dp_timeo_t *dt;
int to;
{
    dt->dt_active =
    dt->dt_active_c =
    dt->dt_active_u =
    dt->dt_waiting =
    dt->dt_failcall = to;
}

static void
dp_maptimeos(dt, from, to)
dp_timeo_t *dt;
int from, to;
{
    if (dt->dt_active == from)
	dt->dt_active = to;
    if (dt->dt_active_c == from)
	dt->dt_active_c = to;
    if (dt->dt_active_u == from)
	dt->dt_active_u = to;
    if (dt->dt_waiting == from)
	dt->dt_waiting = to;
    if (dt->dt_failcall == from)
	dt->dt_failcall = to;
}

static void
dp_addtimeos(dt0, dt1)
dp_timeo_t *dt0,
	   *dt1;
{
    if (dt0->dt_active == DPT_UNSPEC_TIMEOUT &&
	dt1->dt_active != DPT_UNSPEC_TIMEOUT)
	dt0->dt_active = dt1->dt_active;
    if (dt0->dt_active_c == DPT_UNSPEC_TIMEOUT &&
	dt1->dt_active_c != DPT_UNSPEC_TIMEOUT)
	dt0->dt_active_c = dt1->dt_active_c;
    if (dt0->dt_active_u == DPT_UNSPEC_TIMEOUT &&
	dt1->dt_active_u != DPT_UNSPEC_TIMEOUT)
	dt0->dt_active_u = dt1->dt_active_u;
    if (dt0->dt_waiting == DPT_UNSPEC_TIMEOUT &&
	dt1->dt_waiting != DPT_UNSPEC_TIMEOUT)
	dt0->dt_waiting = dt1->dt_waiting;
    if (dt0->dt_failcall == DPT_UNSPEC_TIMEOUT &&
	dt1->dt_failcall != DPT_UNSPEC_TIMEOUT)
	dt0->dt_failcall = dt1->dt_failcall;
}

static int
commentline(b)
char *b;
{
    while (isspace(*b))
	b++;
    return *b == '#';
}

static char *
agetln(b, n)
char *b;
int n;
{
    char *fg, *p;
    if (!access_fp)
	return NULL;
    do {
	fg = fgets(b, n, access_fp);
	if (!fg)
	    return fg;
	if (!(p = strchr(b, '\n'))) {
	    yyerror(access_file, "Line too long");
	    return (char *)0;
	}
	*p = '\0';
	linenum++;
    } while (commentline(b));
    return fg;
}

static void
askipln()
{
    char lnbuf[MAXLN];
    (void)agetln(lnbuf, sizeof lnbuf);
}

static char *put_back_tok;

static void
aputbacktok(t)
char *t;
{
    put_back_tok = t;
}

static char *
agettok()
{
    static char *space_sep;
    static char *ln;
    char *tp;
    static char lnbuf[MAXLN];
    static char tokbuf[MAXLN];

    if (put_back_tok) {
	tp = put_back_tok;
	put_back_tok = (char *)0;
	return tp;
    }

    if (!ln) {
nextln:
	ln = agetln(lnbuf, sizeof lnbuf);
	if (!ln)
	    return (char *)0;
    }
    while (*ln && strchr(SPACES, *ln))
	ln++;
    if (!*ln || *ln == '#')
	goto nextln;
    tp = tokbuf;
    if (strchr(SPECIAL, *ln)) {
	*tp++ = *ln++;
	if ((ln[-1] == '|' || ln[-1] == '&') && ln[-1] == ln[0])
	    *tp++ = *ln++;
	*tp = '\0';
	return tokbuf;
    }
    else {
	for ( ; ; ) {
	    if (!*ln)
		ln = (char *)0;
	    if (!ln || strchr(SPACES, *ln) || strchr(SPECIAL, *ln)) {
		*tp = '\0';
		return tokbuf;
	    }
	    *tp++ = *ln++;
	}
    }
}

static int
parse_access(ap, vers)
access_t *ap;
int vers;
{
    int stat;
    switch (vers) {
     case 1:
	stat = parse_access_v1(ap);
	break;
     case 2:
	stat = parse_access_v2(ap);
	break;
    }
    add_default_condition(ap->Timeouts);
    add_default_condition(ap->DTimeouts);
    add_default_condition(ap->ATimeouts);
    return stat;
}

int
readaccess(cpath, path, rp)
char *cpath,
     *path;
REMOTE *rp;
{
    int i;
    FILE *F = (FILE *)0;
    char *p, *word;
    char buff[MAXLN];
    dp_timeo_t *timeouts;
    access_t *ap;
    char *lasttok;
    int version, subversion;
    int stat;

    if (*path == '\0')
	return errorcount == 0;

    path = expand_dirs_file("$DPCONF_DIR:$DPACCESS_DIR", path);

    /*
     * Look to see if we have parsed this file already.
     */
    for (ap = AccessList ; ap ; ap = ap->Next) {
	if (strcmp(ap->Path, path) == 0) {
	    rp->Access = ap;
	    (void)free(path);
	    return errorcount == 0;
	}
    }

    /*
     * A new access file..  make sure file is there..
     */
    if ((access_fp = fopen(path, "r")) == NULL) {
	yyerror(cpath, "Can't open access file \"%s\"", path);
	(void)free(path);
	return 0;
    }
    access_file = path;

    ap = alloc_access_file(path);
    if (!ap) {
	yyerror(cpath, "Can't allocate memory for  access file \"%s\"", path);
	(void)free(path);
	return 0;
    }
    rp->Access = ap;

    linenum = 0;
    buff[0] = '\0';
    if (!agetln(buff, sizeof buff)) {
	if (!buff[0])
	    yyerror(access_file, "Empty Access File");
	goto badaccessfile;
    }
    rewind(access_fp);
    if (strncmp("DPaccess ", buff, sizeof("DPaccess ")-1) == 0) {
	p = buff + sizeof("DPaccess ") - 1;
	if (!isdigit(*p)) {
	    yyerror(access_file, "Bad Access File version \"%s\"", p);
	    goto badaccessfile;
	}
	version = atoi(p);
	while (isdigit(*p))
	    p++;
	if (*p == '.') {
	    p++;
	    subversion = atoi(p);
	    while (isdigit(*p))
		p++;
	}
	if (*p) {
	    yyerror(access_file, "Bad Access File version \"%s\"",
		    buff + sizeof("DPaccess ") - 1);
	    goto badaccessfile;
	}
    }
    else {
	version = 1;
	subversion = 0;
    }
    linenum = 0;
	
    switch (version) {
     case 1:
     case 2:
	stat = parse_access(ap, version);
	break;
     default:
	yyerror(access_file, "Unknown Access File version \"%s\"", p);
	stat = 0;
	break;
    }
    (void)fclose(access_fp);
    access_fp = (FILE *)0;
    if (stat)
	return stat;

badaccessfile:
    if (access_fp)
	(void)fclose(access_fp);
    access_fp = (FILE *)0;
    (void)free(path);
    AccessList = ap->Next;
    free_access_file(ap);
    return 0;
}

static char **
tok_vector(t, sep, tcp)
char *t;
int sep;
int *tcp;
{
    char **tv0, **tv;
    int tc;
    int n;
    char *cp, *ce;

    for ( n = 1, cp = t ; cp = strchr(cp, sep) ; n++)
	cp++;
    
    if (tcp)
	*tcp = n;

    tv0 = tv = (char **)malloc(n * sizeof(*tv));
    
    cp = t;
    do {
	if (ce = strchr(cp, ':'))
	    n = ce - cp;
	else
	    n = strlen(cp);
	*tv = malloc((unsigned)n+1);
	strncpy(*tv, cp, n);
	(*tv)[n] = '\0';
	tv++;
	cp = ce ? ce + 1 : ce;
    } while (cp);
    *tv = (char *)0;

    return tv0;
}

static void
free_tok_vec(tp)
char **tp;
{
    char **t;
    for (t = tp ; *t ; t++)
	(void)free(*t);
    (void)free((char *)tp);
}


static void
rm_tok_from_vec(tv, tcp, t)
char **tv;
int *tcp;
int t;
{
    int tc = *tcp;
    free(tv[t]);
    tc--;
    if (t < tc)
	memcpy(&tv[t], &tv[t+1], (tc - t)*sizeof(*tv));
    tv[tc] = (char *)0;
    *tcp = tc;
}

static int
qual_in_tab(tv, tcp, tab)
char **tv;
int *tcp;
mapping_t *tab;
{
    int tc = *tcp, t;
    int v;
    for (t = 1 ; t < tc ; t++) {
	if ((v = mapname2num(tv[t], tab)) >= 0) {
	    rm_tok_from_vec(tv, tcp, t);
	    return v;
	}
    }
    return 0;
}

static int
parse_access_v2(ap)
access_t *ap;
{
    char *directive, *qualifier;
    char *token;
    char **tv;
    int tc, src_dst, disallow, which;
    dp_timeo_t *timeouts;
    char *space_sep;
    int n;


    /*
     * Skip over file version line
     */
    askipln();

    while (token = agettok()) {
	tv = tok_vector(token, ':', &tc);
	directive = tv[0];

	if (EQ(directive, "protos")) {
	    src_dst = qual_in_tab(tv, &tc, qual_tab);
	    disallow = qual_in_tab(tv, &tc, allow_tab);
	    if (tc > 1)
		goto unknown_dir;
	    if (ap->Protos)
		yyerror(access_file,
			"permitted protocols can only be specified once");
	    ap->Protos = parse_condition_cbrk(COND_PROTO|src_dst, disallow);
	}
	else if (EQ(directive, "addrs")) {
	    src_dst = qual_in_tab(tv, &tc, qual_tab);
	    disallow = qual_in_tab(tv, &tc, allow_tab);
	    if (tc > 1)
		goto unknown_dir;
	    if (ap->Addrs)
		yyerror(access_file,
			"permitted addresses can only be specified once");
	    ap->Addrs = parse_condition_cbrk(COND_ADDR|src_dst, disallow);
	}
	else if (EQ(directive, "times")) {
	    switch (qual_in_tab(tv, &tc, time_tab)) {
	     case TIMES_CALL:
		disallow = qual_in_tab(tv, &tc, allow_tab);
		if (tc > 1)
		    goto unknown_dir;
		if (ap->CallTimes)
		    yyerror(access_file,
			    "permitted times:call can only be specified once");
		ap->CallTimes = parse_condition_cbrk(COND_TIME, disallow);
		break;
	     case TIMES_CONNECT:
		disallow = qual_in_tab(tv, &tc, allow_tab);
		if (tc > 1)
		    goto unknown_dir;
		if (ap->ConnectTimes)
		    yyerror(access_file,
			    "times:connect can only be specified once");
		ap->ConnectTimes = parse_condition_cbrk(COND_TIME, disallow);
		break;
	     default:
		    goto unknown_dir;
	    }
	}
	else if (EQ(directive, "timeouts")) {
	    which = qual_in_tab(tv, &tc, timeout_tab);
	    if (tc > 1)
		goto unknown_dir;
	    switch (which) {
	     case TIMEOUT_BOTH:
#if	0
		if (ap->Timeouts)
		    yyerror(access_file, "timeouts can only be specified once");
#endif
		parse_timeouts(&ap->Timeouts);
		break;
	     case TIMEOUT_DIAL:
#if	0
		if (ap->DTimeouts)
		    yyerror(access_file,
			    "timeouts:dial can only be specified once");
#endif
		parse_timeouts(&ap->DTimeouts);
		break;
	     case TIMEOUT_ANSWER:
#if	0
		if (ap->ATimeouts)
		    yyerror(access_file, 
			    "timeouts:answer can only be specified once");
#endif
		parse_timeouts(&ap->ATimeouts);
		break;
	     default:
		    goto unknown_dir;
	    }
	}
	else
unknown_dir:
	    yyerror(access_file, "Unknown directive \"%s\"", token);
	
	free_tok_vec(tv);
    }
    return errorcount == 0;
}

static int
parse_seconds_sub(s0, sp)
char *s0,
     **sp;
{
    char *s = *sp - 1;
    while (s >= s0 && isdigit(*s))
	s--;
    if (s > s0 && *s != ':')
	yyerror(access_file, "Bad timeout value \"%s\"", s0);
    *sp = s;
    return atoi(s + 1);
}

static int
parse_seconds(s0)
char *s0;
{
    char *s;
    int sec;
    if (EQ(s0, "none"))
	return DPT_NO_TIMEOUT;
    else if (EQ(s0, "default"))
	return DPT_DEF_TIMEOUT;
    else {
	if (ispositivenumber(s0))
	    return atoi(s0);
	else if (strchr(s0, ':')) {
	    s = s0 + strlen(s0);
	    sec = parse_seconds_sub(s0, &s);
	    if (s > s0)
		sec += parse_seconds_sub(s0, &s) * 60;
	    if (s > s0)
		sec += parse_seconds_sub(s0, &s) * 60 * 60;
	    if (s > s0)
		sec += parse_seconds_sub(s0, &s) * 60 * 60 * 24;
	    if (s > s0)
		goto bad_timeout_val;
	}
	else
bad_timeout_val:
	    yyerror(access_file, "Bad timeout value \"%s\"", s0);
    }
    return sec;
}

static void
parse_timeout(dt, which)
dp_timeo_t *dt;
int which;
{
    char *val = agettok();

    if (!val) {
	yyerror(access_file, "Unexpected EOF while parsing timeout");
	return;
    }
    if (EQ(val, "}")) {
	yyerror(access_file, "Missing timeout value");
	return;
    }

    switch (which) {
     case TO_FAILCALL:	dt->dt_failcall = parse_seconds(val);	break;
     case TO_WAITING:	dt->dt_waiting  = parse_seconds(val);	break;
     case TO_ACTIVE:	dt->dt_active   = parse_seconds(val);	break;
     case TO_ACTIVE_C:	dt->dt_active_c = parse_seconds(val);	break;
     case TO_ACTIVE_U:	dt->dt_active_u = parse_seconds(val);	break;
     default:		yyerror(access_file, "Bad timeout designation");
			break;
    }
}

static mapping_t *
getptabbyname(name, tab)
char *name;
mapping_t *tab;
{
    for ( ; tab->Name ; tab++) {
	if (EQ(tab->Name, name))
	   return tab;
    }
    return (mapping_t *)0;
}

static mapping_t *
getptabbynum(num, tab)
long num;
mapping_t *tab;
{
    for ( ; tab->Name ; tab++) {
	if (tab->Number == num)
	   return tab;
    }
    return (mapping_t *)0;
}

static long
mapname2num(p, tab)
char *p;
mapping_t *tab;
{
    mapping_t *pt;
    return (pt = getptabbyname(p, tab)) ? pt->Number : -1;
}

static char *
mapnum2name(p, tab)
long p;
mapping_t *tab;
{
    mapping_t *pt;
    return (pt = getptabbynum(p, tab)) ? pt->Name : (char *)0;
}

/*
 * Match_pat returns the logical OR of a match of each range in the list.
 */
static int
match_rangelist(pr, r)
long pr;
range_t *r;
{
    for ( ; r ; r = r->Next )
	if (pr >= r->Min && pr <= r->Max)
	    return 1;
    return 0;
}

#ifdef	DEBUG
#ifdef	__STDC__
char *cond_str(condition_t *);
#else
char *cond_str(/* condition_t * */);
#endif
#endif

/*
 * Tables for interpreting protocols symbolically
 */
static long ip_proto_levs[1]        = { PPP_IP };
static long iptcp_proto_levs[2]     = { PPP_IP, IPPROTO_TCP };
static long ipudp_proto_levs[2]     = { PPP_IP, IPPROTO_UDP };
static long ipicmp_proto_levs[2]    = { PPP_IP, IPPROTO_ICMP };
static long ipicmp_unreach_levs[3]  = { PPP_IP, IPPROTO_ICMP, ICMP_UNREACH };
static long ipicmp_redirect_levs[3] = { PPP_IP, IPPROTO_ICMP, ICMP_REDIRECT };
static long ipicmp_timxceed_levs[3] = { PPP_IP, IPPROTO_ICMP, ICMP_REDIRECT };

static path_t ip_proto         = { COND_PROTO, 1, ip_proto_levs };
static path_t iptcp_proto      = { COND_PROTO, 2, iptcp_proto_levs };
static path_t ipudp_proto      = { COND_PROTO, 2, ipudp_proto_levs };
static path_t ipicmp_proto     = { COND_PROTO, 2, ipicmp_proto_levs };
static path_t ipicmp_unreach   = { COND_PROTO, 3, ipicmp_unreach_levs };
static path_t ipicmp_redirect  = { COND_PROTO, 3, ipicmp_redirect_levs };
static path_t ipicmp_timxceed  = { COND_PROTO, 3, ipicmp_timxceed_levs };

static mapping_t ppp_ptab[] = {
#ifdef	PPP_IP	
    { PPP_IP, "ip" },
#endif
    { 0, (char *)0 }
};

static mapping_t icmp_ttab[] = {
#ifdef	ICMP_ECHOREPLY
    { ICMP_ECHOREPLY,		"echo_reply" },
#endif
#ifdef	ICMP_UNREACH
    { ICMP_UNREACH,		"unreach" },
#endif
#ifdef	ICMP_SOURCEQUENCH
    { ICMP_SOURCEQUENCH,	"source_quench" },
#endif
#ifdef	ICMP_REDIRECT
    { ICMP_REDIRECT,		"redirect" },
#endif
#ifdef	ICMP_ECHO
    { ICMP_ECHO,		"echo" },
#endif
#ifdef	ICMP_TIMXCEED
    { ICMP_TIMXCEED,		"time_exceeded" },
#endif
#ifdef	ICMP_PARAMPROB
    { ICMP_PARAMPROB,		"param_problem" },
#endif
#ifdef	ICMP_TSTAMP
    { ICMP_TSTAMP,		"timestamp_request" },
#endif
#ifdef	ICMP_TSTAMPREPLY
    { ICMP_TSTAMPREPLY,		"timestamp_reply" },
#endif
#ifdef	ICMP_IREQ
    { ICMP_IREQ,		"info_request" },
#endif
#ifdef	ICMP_IREQREPLY
    { ICMP_IREQREPLY,		"info_reply" },
#endif
#ifdef	ICMP_MASKREQ
    { ICMP_MASKREQ,		"addr_mask_request" },
#endif
#ifdef	ICMP_MASKREPLY
    { ICMP_MASKREPLY,		"addr_mask_reply" },
#endif
    { 0, (char *)0 }
};

static mapping_t icmp_unreach_ctab[] = {
#ifdef	ICMP_UNREACH_NET
    { ICMP_UNREACH_NET,		"net" },
#endif
#ifdef	ICMP_UNREACH_HOST
    { ICMP_UNREACH_HOST,	"host" },
#endif
#ifdef	ICMP_UNREACH_PROTOCOL
    { ICMP_UNREACH_PROTOCOL,	"protocol" },
#endif
#ifdef	ICMP_UNREACH_PORT
    { ICMP_UNREACH_PORT,	"port" },
#endif
#ifdef	ICMP_UNREACH_NEEDFRAG
    { ICMP_UNREACH_NEEDFRAG,	"needfrag" },
#endif
#ifdef	ICMP_UNREACH_SRCFAIL
    { ICMP_UNREACH_NET,		"srcfail" },
#endif
    { 0, (char *)0 }
};

static mapping_t icmp_redirect_ctab[] = {
#ifdef	ICMP_REDIRECT_NET
    { ICMP_REDIRECT_NET,	"net" },
#endif
#ifdef	ICMP_REDIRECT_HOST
    { ICMP_REDIRECT_HOST,	"host" },
#endif
#ifdef	ICMP_REDIRECT_TOSNET
    { ICMP_REDIRECT_TOSNET,	"tosnet" },
#endif
#ifdef	ICMP_REDIRECT_TOSHOST
    { ICMP_REDIRECT_TOSHOST,	"toshost" },
#endif
    { 0, (char *)0 }
};

static mapping_t icmp_timxceed_ctab[] = {
#ifdef	ICMP_TIMXCEED_NET
    { ICMP_TIMXCEED_NET,	"intrans" },
#endif
#ifdef	ICMP_TIMXCEED_HOST
    { ICMP_TIMXCEED_HOST,	"reass" },
#endif
    { 0, (char *)0 }
};

/*
 * These functions know a little about some networks, especially IP,
 * so that more meaningful things can be printed.
 */
char *
ppath_str(p)
path_t *p;
{
    static char sbuf[64];
    char *s = sbuf, *ps;
    char *pr;
    int lev;
    int val;

    for (lev = 0 ; lev < p->NLevels ; lev++) {
	val = p->Levels[lev];
	switch (lev) {
	 case 0:
	    if (ps = mapnum2name(val, ppp_ptab)) {
		(void)strcpy(s, ps);
		break;
	    }
	    goto num;
	 case 1:
	    if (p->Levels[0] == PPP_IP) {
		struct protoent *pe;

		if (pe = getprotobynumber(val)) {
		    (void)strcpy(s, pe->p_name);
		    break;
		}
	    }
	    goto num;
	 case 2:
	    if (p->Levels[0] == PPP_IP) {
		switch (p->Levels[1]) {
		 case IPPROTO_UDP:
		    pr = "udp";
		    goto serv;
		 case IPPROTO_TCP:
		    pr = "tcp";
		  serv:
		    {
			struct servent *se;
			if (se = getservbyport(val, pr)) {
			    (void)strcpy(s, se->s_name);
			    break;
			}
		    }
		    goto num;
		 case IPPROTO_ICMP:
		    if (ps = mapnum2name(val, icmp_ttab)) {
			(void)strcpy(s, ps);
			break;
		    }
		    goto num;
		 default:
		    goto num;
		}
		break;
	    }
	    goto num;
	 case 3:
	    if (p->Levels[0] == PPP_IP && p->Levels[1] == IPPROTO_ICMP) {
		switch (p->Levels[2]) {
		 case ICMP_UNREACH:
		    if (ps = mapnum2name(val, icmp_unreach_ctab)) {
			(void)strcpy(s, ps);
			break;
		    }
		    goto num;
		 case ICMP_REDIRECT:
		    if (ps = mapnum2name(val, icmp_redirect_ctab)) {
			(void)strcpy(s, ps);
			break;
		    }
		    goto num;
		 case ICMP_TIMXCEED:
		    if (ps = mapnum2name(val, icmp_timxceed_ctab)) {
			(void)strcpy(s, ps);
			break;
		    }
		    goto num;
		}
	    }
	    goto num;
	num:
	 default:
	    (void)sprintf(s, "%d", p->Levels[lev]);
	    break;
	}
	s += strlen(s);
	if (lev < p->NLevels - 1)
	    *s++ = '/';
    }
    *s = '\0';
    return sbuf;
}

char *
apath_str(p)
path_t *p;
{
    static char sbuf[64];
    char *s = sbuf, *ps;
    int lev;

    for (lev = 0 ; lev < p->NLevels ; lev++) {
	switch (lev) {
	 case 0:
	    if (ps = mapnum2name(p->Levels[lev], ppp_ptab)) {
		(void)strcpy(s, ps);
		break;
	    }
	 default:
	    (void)sprintf(s, "%d", p->Levels[lev]);
	    break;
	}

	s += strlen(s);
	if (lev < p->NLevels - 1)
	    *s++ = lev > 0 ? '.' : '/';
    }
    *s = '\0';
    return sbuf;
}

char *
tpath_str(p)
path_t *p;
{
    static char sbuf[64];
    char *s = sbuf;
    int lev;

    for (lev = 0 ; lev < p->NLevels ; lev++) {
	(void)strcpy(s, time_str(p->Levels[lev], lev));
	s += strlen(s);
	if (lev < p->NLevels - 1)
	    *s++ = '/';
    }
    *s = '\0';
    return sbuf;
}

char *
time_str(t, lev)
long t;
int lev;
{
    static char tbuf[32];
    switch (lev) {
     case 0:
	switch (t) {
	 case 0:	return "sun";
	 case 1:	return "mon";
	 case 2:	return "tue";
	 case 3:	return "wed";
	 case 4:	return "thu";
	 case 5:	return "fri";
	 case 6:	return "sat";
	 default:	(void)sprintf(tbuf, "%d", t); break;
	}
	break;
     case 1:
	(void)sprintf(tbuf, "%02d:%02d:%02d",
		      t / (60*60), t % (60*60) / 60, t % 60);
	break;
     default:
	(void)sprintf(tbuf, "%d", t);
	break;
    }
    return tbuf;
}

char *
path_str(p)
path_t *p;
{
    if (!p)
	return "none";
    switch (p->Type) {
     case COND_PROTO:
     case COND_DSTPROTO:
     case COND_SRCPROTO:
	return ppath_str(p);
     case COND_ADDR:
     case COND_DSTADDR:
     case COND_SRCADDR:
	return apath_str(p);
     case COND_TIME:
	return tpath_str(p);
     default:
	return "Unrecognized path type";
    }
}

char *
appath_str(ap, pp)
path_t *ap, *pp;
{
    static char apbuf[128];
    char *a, *p, *p1;

    a = apath_str(ap);
    p = ppath_str(pp);

    /*
     * If the address family is the same as the protocol family,
     * don't report it twice.
     */
    if ((p1 = strchr(p, '/')) &&
	(strncmp(a, p, p1 - p + 1) == 0))
	p = p1 + 1;

    (void)sprintf(apbuf, "%s:%s", a, p);
    return apbuf;
}

char *
addr_proto_str(pi)
pkt_info_t *pi;
{
    static char apbuf[128];
    char *addr, *proto;
    char *p = apbuf;

    (void)sprintf(p, "%s -> ", appath_str(pi->SrcAddr, pi->SrcProto));
    p += strlen(p);
    (void)sprintf(p, "%s",     appath_str(pi->DstAddr, pi->DstProto));
    return apbuf;
}

/*
 * Match_pat returns the logical AND of a match at each level of the protocol.
 */
static int
match_pat(pr, c)
path_t *pr;
condition_t *c;
{
    int lev;
    pattern_t *pat = c->Pattern;
#ifdef	DEBUG
    printf("match path %s pattern %s ", path_str(pr), cond_str(c));
#endif
    if (!pr) {
#ifdef	DEBUG
	printf("Failure\n");
#endif
	return 0;
    }
    for (lev = 0 ; pat ; pat = pat->NextLevel, lev++) {
	if (lev >= pr->NLevels) {
#ifdef	DEBUG
	    printf("Failure\n");
#endif
	    return 0;
	}
	if (!match_rangelist(pr->Levels[lev], pat->RangeList)) {
#ifdef	DEBUG
	    printf("Failure\n");
#endif
	    return 0;
	}
    }
#ifdef	DEBUG
	    printf("Success\n");
#endif
    return 1;
}

static char *
lcase(us)
char *us;
{
    char *ls0 = malloc((unsigned)strlen(us)+1);
    char *ls;
    for ( ls = ls0 ; *us ; us++)
	*ls++ = isupper(*us) ? tolower(*us) : *us;
    *ls = '\0';
    return ls0;
}


#define	MAP_MIN		1
#define	MAP_MAX		2
#define	MAP_RANGE	3

static void
set_range(r, which, val)
range_t *r;
int which;
long val;
{
    switch (which) {
     case MAP_MIN:	r->Min = val;		break;
     case MAP_MAX:	r->Max = val;		break;
     case MAP_RANGE:	r->Min = r->Max = val;	break;
    }
}

static void
map_proto_element(r, which, s, level, cond)
range_t *r;
int which;
char *s;
int level;
condition_t *cond;
{
    long p;
    if (ispositivenumber(s)) {
	set_range(r, which, strtol(s, (char *)0, 0));
	return;
    }
    switch (level) {
     case 0:
	p = mapname2num(s, ppp_ptab);
	if (p >= 0) {
	    set_range(r, which, p);
	    return;
	}
	break;
     case 1:
	if (match_pat(&ip_proto, cond)) {
	    struct protoent *pe;
	    char *ls;

	    ls = lcase(s);
	    pe = getprotobyname(ls);
	    free(ls);
	    if (pe) {
		set_range(r, which, pe->p_proto);
		return;
	    }
	}
	break;
     case 2:
	if (match_pat(&iptcp_proto, cond)) {
	    struct servent *se;
	    char *ls;

	    ls = lcase(s);
	    se = getservbyname(ls, "tcp");
	    free(ls);
	    if (se) {
		set_range(r, which, se->s_port);
		return;
	    }
	}
	if (match_pat(&ipudp_proto, cond)) {
	    struct servent *se;
	    char *ls;

	    ls = lcase(s);
	    se = getservbyname(ls, "udp");
	    free(ls);
	    if (se) {
		set_range(r, which, se->s_port);
		return;
	    }
	}
	if (match_pat(&ipicmp_proto, cond)) {
	    int t;
	    if (t = mapname2num(s, icmp_ttab)) {
		set_range(r, which, t);
		return;
	    }
	}
	break;
     case 3:
	if (match_pat(&ipicmp_unreach, cond)) {
	    int t;
	    if (t = mapname2num(s, icmp_unreach_ctab)) {
		set_range(r, which, t);
		return;
	    }
	}
	if (match_pat(&ipicmp_redirect, cond)) {
	    int t;
	    if (t = mapname2num(s, icmp_redirect_ctab)) {
		set_range(r, which, t);
		return;
	    }
	}
	if (match_pat(&ipicmp_timxceed, cond)) {
	    int t;
	    if (t = mapname2num(s, icmp_timxceed_ctab)) {
		set_range(r, which, t);
		return;
	    }
	}
	break;
    }
    yyerror(access_file, "Bad proto element \"%s\"", s);
}

static void
map_addr_element(r, which, s, level, cond)
range_t *r;
int which;
char *s;
int level;
condition_t *cond;
{
    long p;
    if (ispositivenumber(s)) {
	set_range(r, which, strtol(s, (char *)0, 0));
	return;
    }
    switch (level) {
     case 0:
	p = mapname2num(s, ppp_ptab);
	if (p >= 0) {
	    set_range(r, which, p);
	    return;
	}
	break;
    }
    yyerror(access_file, "Bad address element \"%s\"", s);
}

static mapping_t day_ptab[] = {
    { 0, "sun" },
    { 0, "sunday" },
    { 1, "mon" },
    { 1, "monday" },
    { 2, "tue" },
    { 2, "tuesday" },
    { 3, "wed" },
    { 3, "wednesday" },
    { 4, "thu" },
    { 4, "thursday" },
    { 5, "fri" },
    { 5, "friday" },
    { 6, "sat" },
    { 6, "saturday" },
    { 0, (char *)0 }
};

static void
map_time_element(r, which, s0, level, cond)
range_t *r;
int which;
char *s0;
int level;
condition_t *cond;
{
    long d, h, m, s, sa, sb;
    char *s1;
    switch (level) {
     case 0:
	/*
	 * Day of the week
	 */
	if (EQ(s0, "weekdays")) {
	    if (which != MAP_RANGE)
		break;
	    r->Min = 1;
	    r->Max = 5;
	    return;
	}
	else if (EQ(s0, "weekends")) {
	    if (which != MAP_RANGE)
		break;
	    r->Min = 6;
	    r->Max = 0;
	    return;
	}
	d = mapname2num(s0, day_ptab);
	if (d >= 0) {
	    set_range(r, which, d);
	    return;
	}
	break;
     case 1:
	/*
	 * Time of the day
	 */
	h = m = s = -1;
	if (EQ(s0, "none")) {
	    sa = sb = -1;
	}
	else {
	    h = strtol(s0, &s1, 10);
	    if (h < 0)
		break;
	    if (*s1) {
		if (*s1 != ':')
		    break;
		m = strtol(++s1, &s1, 10);
		if (m < 0)
		    break;
		if (*s1) {
		    if (*s1 != ':')
			break;
		    s = strtol(++s1, &s1, 10);
		    if (s < 0)
			break;
		    if (*s1)
			break;
		}
	    }
	    if (h >= 24)
		break;
	    if (m >= 60)
		break;
	    if (s >= 60)
		break;
	    if (m < 0) {
		sa = h * 60*60;
		sb = sa + 60 * 60 - 1;
	    }
	    else if (s < 0) {
		sa = h * 60*60 + m * 60;
		sb = sa + 60 - 1;
	    }
	    else
		sa = sb = h * 60*60 + m * 60 + s;
	}
	switch (which) {
	 case MAP_MIN:	r->Min = sa; 			break;
	 case MAP_MAX:	r->Max = sb; 			break;
	 case MAP_RANGE:r->Min = sa; r->Max = sb;	break;
	}
	return;
    }
    yyerror(access_file, "Bad time element \"%s\"", s0);
}

static void
expand_time_list(tr, lev)
range_t *tr;
int lev;
{
    range_t *ntr;
    int min, max;

    switch (lev) {
     case 0:	min = 0; max = 6;		break;
     case 1:	min = 0; max = 24*60*60-1;	break;
     default:	return;
    }

    while (tr) {
	if (tr->Min > tr->Max) {
	    ntr = alloc_range();
	    ntr->Next = tr->Next;
	    tr->Next = ntr;

	    ntr->Max = tr->Max;
	    tr->Max = max;
	    ntr->Min = min;

	    tr = ntr->Next;
	}
	else
	    tr = tr->Next;
    }
}


static range_t *
parse_range(rs, mapfunc, level, cond)
char *rs;
void (* mapfunc)();
int level;
condition_t *cond;
{
    range_t *r = alloc_range();
    char *re;

    if (*rs == '[') {
	rs++;
	if (!(re = strchr(rs, ',')))
	    goto range_err;
	*re++ = '\0';
	(*mapfunc)(r, MAP_MIN, rs, level, cond);
	rs = re;
	if (!(re = strchr(rs, ']')))
	    goto range_err;
	if (*(re+1))
	    goto range_err;
	*re = '\0';
	(*mapfunc)(r, MAP_MAX, rs, level, cond);
    }
    else {
	(*mapfunc)(r, MAP_RANGE, rs, level, cond);
    }
    return r;

range_err:
    yyerror(access_file, "Bad range \"%s\"", rs);
    free_range(r);
    return (range_t *)0;
}

static char *
next_range(rp)
char **rp;
{
    char *r, *r0;
    if (!*rp)
	return (char *)0;
	
    r0 = *rp;
    if (*r0 == '[') {
	r = strchr(r0, ']');
	if (!r) {
	    yyerror(access_file, "Bad range at \"%s\"", r0);
	    return (char *)0;
	}
	r++;
	if (*r && *r != ',') {
	    yyerror(access_file, "Bad range at \"%s\"", r0);
	    return (char *)0;
	}
    }
    else {
	r = strchr(r0, ',');
    }
    if (r && *r) {
	*r = '\0';
	*rp = r + 1;
    }
    else
	*rp = (char *)0;

    return r0;
}

static char *
first_range(r0, rp)
char *r0,
     **rp;
{
    *rp = r0;
    return next_range(rp);
}

static range_t *
parse_range_list(rls, mapfunc, level, cond)
char *rls;
void (* mapfunc)();
int level;
condition_t *cond;
{
    range_t *r0, *r, **rp = &r0;
    char *rs;
    char *range_sep;

    rs = first_range(rls, &range_sep);
    do {
	if (!(r = parse_range(rs, mapfunc, level, cond)))
	    break;
	*rp = r;
	rp = &r->Next;
    } while (rs = next_range(&range_sep));
    return r0;
}

static condition_t *
parse_proto_path(ppath, path_type)
char *ppath;
int path_type;
{
    condition_t *c;
    char *rls;
    char *slash_sep;
    pattern_t **patp, *p;
    int lev = 0;

    c = alloc_condition(path_type);
    patp = &c->Pattern;

    rls = strtok_r(ppath, SLASH, &slash_sep);
    do {
	p = alloc_pattern();

	p->RangeList = parse_range_list(rls,
					map_proto_element, lev, c);

	*patp = p;
	patp = &p->NextLevel;

	lev++;
    } while (rls = nexttok(SLASH, &slash_sep));
    return c;
}

static pattern_t *
inet_path(cp, n)
char *cp;
int n;
{
    pattern_t *p0, *p, **pp;

    p0 = alloc_pattern();
    set_range(p0->RangeList = alloc_range(), MAP_RANGE, PPP_IP);
    pp = &p0->NextLevel;

    while (n--) {
	p = alloc_pattern();
	set_range(p->RangeList = alloc_range(), MAP_RANGE, *cp++ & 0xff);
	*pp = p;
	pp = &p->NextLevel;
    }

    return p0;
}


static condition_t *
addr2path(hn, path_type)
char *hn;
int path_type;
{
    condition_t *c, *c0, **cp;
    struct hostent *he;
    int n;
    char **ap;
    he = gethostbyname(hn);
    if (!he) {
	yyerror(access_file, "Bad address \"%s\"", hn);
	return (condition_t *)0;
    }
    switch (he->h_addrtype) {
     default:
	yyerror(access_file, "Unknown address family %d for \"%s\"",
		he->h_addrtype, hn);
	return (condition_t *)0;
     case AF_INET:
	c0 = (condition_t *)0;
	cp = &c0;
	for (ap = he->h_addr_list ; *ap ; ap++) {
	    if (*cp) {
		c = alloc_condition(COND_OR);
		c->Left = *cp;
		*cp = c;
		cp = &c->Right;
	    }
	    c =  alloc_condition(path_type);
	    *cp = c;
	    c->Pattern = inet_path(*ap, he->h_length);
	}
	return c0;
    }
}

static condition_t *
parse_addr_path(ppath, path_type)
char *ppath;
int path_type;
{
    condition_t *c;
    char *rls;
    char *slash_sep;
    char *sep = SLASH;
    pattern_t **patp, *p;
    int lev = 0;

    if (!strchr(ppath, '/') &&
	!getptabbyname(ppath, ppp_ptab) &&
	isinetname(ppath))
	return addr2path(ppath, path_type);

    c = alloc_condition(path_type);
    patp = &c->Pattern;

    rls = strtok_r(ppath, sep, &slash_sep);
    do {
	p = alloc_pattern();

	p->RangeList = parse_range_list(rls,
					map_addr_element, lev, c);

	*patp = p;
	patp = &p->NextLevel;

	if (lev == 0 && match_pat(&ip_proto, c))
	    sep = DOTSLASH;

	lev++;
    } while (rls = nexttok(sep, &slash_sep));
    return c;
}

static condition_t *
parse_condition_expr(type, end_tok)
int type,
    end_tok;
{
    char *tok;
    int t;
    condition_t *c = (condition_t *)0, *cop;

    tok = agettok();
    t = mapname2num(tok, op_tab);

    switch (t) {
     case OP_NOT:
	cop = alloc_condition(COND_NOT);
	cop->Right = parse_condition_expr(type, end_tok);
	return cop;
     case OP_LPAR:
	c = parse_condition_expr(type, OP_RPAR);
	goto look_for_op;
     case OP_TRUE:
     case OP_FALSE:
	c = alloc_condition(t);
	goto look_for_op;
     case -1: /* Non keyword */
	if (!(c = parse_path(tok, type)))
	    return (condition_t *)0;
look_for_op:
	tok = agettok();
	t = mapname2num(tok, op_tab);
	if (t < 0)
	    goto syntax_err;
	else if (t == end_tok)
	    return c;
	else if (REL_OP(t)) {
	    cop = alloc_condition(t);
	    cop->Left = c;
	    cop->Right = parse_condition_expr(type, end_tok);
	    return cop;
	}
	goto syntax_err;
     default: /* Other recognized keyword/token */
	if (t == end_tok)
	    return c;
syntax_err:
	yyerror(access_file, "Syntax error near \"%s\"", tok);
	return (condition_t *)0;
    }
}

static condition_t *
parse_condition_cbrk(type, negate)
int type;
int negate;
{
    condition_t *c, *nc;
    char *tok;

    if (!(tok = agettok()) || !EQ(tok, "{")) /* } */
	yyerror(access_file, "looking for '{', got \"%s\"",
		tok ? tok : "EOF"); /* } */
    
    c = parse_condition_expr(type, OP_RCBK);
    if (!c)
	return alloc_condition(negate ? COND_TRUE : COND_FALSE);
    if (negate) {
	nc = alloc_condition(COND_NOT);
	nc->Right = c;
	return nc;
    }
    else
	return c;
}

static condition_t *
parse_time_path(tpath)
char *tpath;
{
    condition_t *c;
    char *rls;
    char *slash_sep;
    pattern_t **patp, *p;
    int lev = 0;

    c = alloc_condition(COND_TIME);
    patp = &c->Pattern;

    rls = strtok_r(tpath, SLASH, &slash_sep);
    do {
	if (lev >= 2) {
	    yyerror(access_file, "Bad time at \"%s\"", rls);
	    break;
	}
	p = alloc_pattern();

	p->RangeList = parse_range_list(rls,
					map_time_element, lev, c);

	expand_time_list(p->RangeList, lev);

	*patp = p;
	patp = &p->NextLevel;

	lev++;
    } while (rls = nexttok(SLASH, &slash_sep));
    return c;
}

static char *
strip_qual(sp)
char **sp;
{
    char *qe,
	 *q;
    q = *sp;
    qe = strchr(q, ':');
    if (!qe)
	return (char *)0;
    *qe++ = '\0';
    *sp = qe;
    return q;
}

static int
prefix_colon(sp, tab)
char **sp;
mapping_t *tab;
{
    char *pe;
    char *s0 = *sp;
    int t;

    pe = strchr(s0, ':');

    if (!pe)
	return -1;
    
    *pe = '\0';

    t = mapname2num(s0, tab);

    if (t < 0)
	*pe = ':';		/* put back */
    else
	*sp = pe + 1;		/* skip over */
    return t;
}

static condition_t *
parse_path(path, type)
char *path;
int type;
{
    condition_t *c;
    char *qual;
    int t;
    if (COND_COND(type) == 0) {
	if ((t = prefix_colon(&path, cond_tab)) >= 0)
	    type |= t;
	else {
	    yyerror(access_file, "Unknown condition type \"%s\"", path);
	    return (condition_t *)0;
	}
    }
    if ((t = prefix_colon(&path, qual_tab)) >= 0)
	type |= t;

    switch (COND_COND(type)) {
     case COND_PROTO:	return parse_proto_path(path, type);
     case COND_ADDR:	return parse_addr_path(path, type);
     case COND_TIME:	return parse_time_path(path);
    }
    yyerror(access_file, "Unknown condition type \"%s\"", path);
}

static void
parse_timeouts(top)
timeout_t **top;
{
    char *space_sep;
    char *tok;
    int t;
    int which;
    timeout_t *to;
    char buff[BUFSIZ];
    condition_list_t *cl;

    if (!(tok = agettok()) || !EQ(tok, "{")) /* } */
	yyerror(access_file, "looking for '{', got \"%s\"",
		tok ? tok : "EOF"); /* } */

    if (!(to = *top))
	to = *top = alloc_timeout();

    while (tok = agettok()) {
	if (EQ(tok, "}")) /* { */
	    return;
	t = mapname2num(tok, to_which_tab);
	if (t > 0)
	    parse_timeout(&to->DefaultCondition->Timeouts, t);
	else {
	    aputbacktok(tok);
	    cl = alloc_condition_list();
	    cl->Conditions = parse_condition_expr(0, OP_LCBK);
	    while (tok = agettok()) {
		if (EQ(tok, "}")) /* { */
		    break;;
		t = mapname2num(tok, to_which_tab);
		if (t <= 0)
		    yyerror(access_file, "Unknown timeout \"%s\"", tok);
		else
		    parse_timeout(&cl->Timeouts, t);
	    }
	    add_condition_list(to, cl);
	    if (!tok)
		goto unexpected_EOF;
	}
    }
unexpected_EOF:
    yyerror(access_file, "Unexpected EOF while parsing timeouts");
}

#define	hexlead(s) (strncmp((s), "0x", 2) == 0 || strncmp((s), "0X", 2) == 0)

static int
ispositivenumber(p)
    char	*p;
{
    if (*p == '\0')
	return 0;
    if (hexlead(p))
	p += 2;
    for ( ; *p ; p++)
	if (!(isdigit(*p)))
	    return 0;
    return 1;
}

static int
isinetname(s)
char *s;
{
    if (!isalpha(*s))
	return 0;
    for (s++ ; *s ; s++) {
	if (*s == '.')
	    continue;
	if (isalpha(*s))
	    continue;
	if (isdigit(*s) && *(s-1) != '.')
	    continue;
	return 0;
    }
    return 1;
}

void
dlsap2ph(ph, sap)
phdr_t *ph;
ulong sap;
{
    ph->ph_address  = DP_DLSAP_ADDR(sap);
    ph->ph_control  = DP_DLSAP_CTL(sap);
    ph->ph_protocol = DP_DLSAP_PROTO(sap);
}

/*
 * These functions know about a lot of protocol specific formats.
 * It extracts several levels of protocol information from the packet.
 */
static path_t *
packet2proto(dlu, pkt, dst_flag)
register dl_unitdata_ind_t *dlu;
char *pkt;
int dst_flag;
{
    long plevs[6];
    int nlevs = 0;
    ulong dlsap;

    dlsap = *(ulong *)(((char *)dlu) + dlu->dl_dest_addr_offset);
    switch (plevs[nlevs++] = DP_DLSAP_PROTO(dlsap) & ~DPP_DIAL_REQ) {
     case PPP_IP:
	{
	    /*
	     * Check IP Protocol (TCP, UDP, etc)
	     */
	    struct ip *ip = (struct ip *)pkt;
	    switch (plevs[nlevs++] = ip->ip_p) {
	     case IPPROTO_TCP:
		{
		    struct tcphdr *th;
		    th = (struct tcphdr *)&((char *)ip)[ip->ip_hl << 2];
		    plevs[nlevs++] = ntohs(dst_flag ? th->th_dport : th->th_sport);
		}
		break;
	     case IPPROTO_UDP:
		{
		    struct udphdr *uh;
		    uh = (struct udphdr *)&((char *)ip)[ip->ip_hl << 2];
		    plevs[nlevs++] = ntohs(dst_flag ? uh->uh_dport : uh->uh_sport);
		}
		break;
	     case IPPROTO_ICMP:
		{
		    struct icmp *icmp;
		    icmp = (struct icmp *)&((char *)ip)[ip->ip_hl << 2];
		    if (dst_flag) {
			plevs[nlevs++] = icmp->icmp_type;
			if (icmp->icmp_code)
			    plevs[nlevs++] = icmp->icmp_code;
		    }
		}
	    }
	}
    }
    return new_path(COND_PROTO, nlevs, plevs);
}

static path_t *
packet2addr(dlu, pkt, dst_flag)
register dl_unitdata_ind_t *dlu;
char *pkt;
int dst_flag;
{
    long plevs[6];
    int nlevs = 0;
    ulong dlsap;

    dlsap = *(ulong *)(((char *)dlu) + dlu->dl_dest_addr_offset);
    switch (plevs[nlevs++] = DP_DLSAP_PROTO(dlsap) & ~DPP_DIAL_REQ) {
     case PPP_IP:
	{
	    /*
	     * Extract appropriate IP address;
	     */
	    struct ip *ip = (struct ip *)pkt;
	    struct in_addr *ia;
	    long addr;

	    ia = dst_flag ? &ip->ip_dst : &ip->ip_src;
	    addr = ntohl(ia->s_addr);

	    plevs[nlevs++] = (addr >> 24) & 0xff;
	    plevs[nlevs++] = (addr >> 16) & 0xff;
	    plevs[nlevs++] = (addr >>  8) & 0xff;
	    plevs[nlevs++] = (addr >>  0) & 0xff;
	}
	break;
    }
    return new_path(COND_PROTO, nlevs, plevs);
}

static path_t *
time_path(time)
time_t time;
{
    struct tm *tm;
    long daysec[2];
    tm = localtime(&time);
    daysec[0] = tm->tm_wday;
    daysec[1] = tm->tm_hour * (60*60)
	      + tm->tm_min * 60
	      + tm->tm_sec;
    return new_path(COND_TIME, 2, daysec);
}

static path_t *
current_time_path()
{
    return time_path(time((time_t *)0));
}

static pkt_info_t *
pkt_info(dlu, pkt)
register dl_unitdata_ind_t *dlu;
char *pkt;
{
    pkt_info_t *pi = alloc_pkt_info();
    if (dlu && pkt) {
	pi->DstProto = packet2proto(dlu, pkt, 1);
	pi->SrcProto = packet2proto(dlu, pkt, 0);
	pi->DstAddr  = packet2addr(dlu, pkt, 1);
	pi->SrcAddr  = packet2addr(dlu, pkt, 0);
    }
    pi->CurrentTime = current_time_path();
    return pi;
}

int
authorized(ap, dlu, dlu_len, pkt, pkt_len)
register access_t *ap;
register dl_unitdata_ind_t *dlu;
int dlu_len;
char *pkt;
int pkt_len;
{
    register int		i;
    struct timeval		tv;
    struct tm			*tm;
    phdr_t			phdr;
    pkt_info_t			*pi;
    static char			WHERE[] = "authorization check";
    char			*err = (char *)0;

    /*
     * If no restrictions, assume call is okay.
     */
    if (!ap)
	return 1;

    /*
     * Check protocols involved
     */
    pi = pkt_info(dlu, pkt);

    if (!match_cond(ap->CallTimes, pi)) {
	err = "time";
    }
    else if (!match_cond(ap->Protos, pi)) {
	err = "protocol";
    }
    else if (!match_cond(ap->Addrs, pi)) {
	err = "address";
    }
	
    if (err) {
	d_log(DLOG_GENERAL, WHERE, "Bad %s for call: %s",
	      err, addr_proto_str(pi));
	free_pkt_info(pi);
	return 0;
    }


    /*
     * Currently, we only support IP.
     */
    if (pi->DstProto->Levels[0] != PPP_IP) {
	dlsap2ph(&phdr, pi->DstProto->Levels[0]);
	d_log(DLOG_INFO, WHERE,
	      "Unknown protocol %02x:%02x:%04x",
	      phdr.ph_address,
	      phdr.ph_control,
	      ntohs(phdr.ph_protocol));
not_authorized:
	free_pkt_info(pi);
	return 0;
    }
    /*
     * Okay to make call
     */
    free_pkt_info(pi);
    return 1;
}

static int
match_cond(c, pi)
condition_t *c;
pkt_info_t *pi;
{
    static char		WHERE[] = "match_cond";
    if (!c)
	return 1;
    if (COND_BRANCH(c->Type)) {
	switch (c->Type) {
	 case COND_OR:
	    return match_cond(c->Left, pi) || match_cond(c->Right, pi);
	 case COND_AND:
	    return match_cond(c->Left, pi) && match_cond(c->Right, pi);
	 case COND_NOT:
	    return !match_cond(c->Right, pi);
	}
    }
    else {
	switch (c->Type) {
	 case COND_DSTPROTO:
	    return match_pat(pi->DstProto, c);
	 case COND_SRCPROTO:
	    return match_pat(pi->SrcProto, c);
	 case COND_PROTO:
	    return match_pat(pi->DstProto, c) || match_pat(pi->SrcProto, c);
	 case COND_DSTADDR:
	    return match_pat(pi->DstAddr, c);
	 case COND_SRCADDR:
	    return match_pat(pi->SrcAddr, c);
	 case COND_ADDR:
	    return match_pat(pi->DstAddr, c) || match_pat(pi->SrcAddr, c);
	 case COND_TIME:
	    return match_pat(pi->CurrentTime, c);
	 case COND_TRUE:
	    return 1;
	 case COND_FALSE:
	    return 0;
	}
    }
    d_log(DLOG_DIAG, WHERE, "Unknown condition type 0x%x", c->Type);
    return 0;
}

static void
timeout_rules(dt, to, pi)
dp_timeo_t *dt;
timeout_t *to;
pkt_info_t *pi;
{
    condition_list_t *cl;
    condition_t *c;

    if (!to)
	return;

    if (!any_timeos(dt, DPT_UNSPEC_TIMEOUT))
	return;

    for (cl = to->ConditionList ; cl ; cl = cl->Next)
	if (match_cond(cl->Conditions, pi)) {
	    dp_addtimeos(dt, &cl->Timeouts);
	    if (!any_timeos(dt, DPT_UNSPEC_TIMEOUT))
		return;
	}
}

void
comp_timeos(dt, ap, dialflag, dlu, pkt)
dp_timeo_t *dt;
access_t *ap;
int dialflag;
register dl_unitdata_ind_t *dlu;
char *pkt;
{
    pkt_info_t *pi;

    pi = pkt_info(dlu, pkt);
    dp_settimeos(dt, DPT_UNSPEC_TIMEOUT);

    timeout_rules(dt, dialflag ? ap->DTimeouts : ap->ATimeouts, pi);
    if (any_timeos(dt, DPT_UNSPEC_TIMEOUT)) {
	timeout_rules(dt, ap->Timeouts, pi);

	if (any_timeos(dt, DPT_UNSPEC_TIMEOUT))
	    dp_maptimeos(dt, DPT_UNSPEC_TIMEOUT, DPT_NO_TIMEOUT);
    }
    free_pkt_info(pi);
}

static int
set_timeouts(rp, dialflag, dlu, pkt)
REMOTE	 *rp;
int dialflag;
register dl_unitdata_ind_t *dlu;
char *pkt;
{
    static char		WHERE[] = "set_timeouts";

    dp_timeo_t dpt;
    
    comp_timeos(&dpt, rp->Access, dialflag, dlu, pkt);

    if (strioctl(rp->DlpiFd, DPIOCSTIMEOS, -1,
		 sizeof(dpt), (char *)&dpt) < 0) {
	d_log(DLOG_GENERAL, WHERE,
	      "Can't set %s timeouts for \"%s\", %m",
	      dialflag ? "Dial" : "Answer", rp->Device);
	return 0;
    }
    d_log(DLOG_ALL, WHERE, "Set %s Timeouts for %s:",
	  dialflag ? "Dial" : "Answer", rp->Device);
    d_log(DLOG_ALL, WHERE,
	  "failedcall %d callwait %d inactivity %d last_close %d non_stream %d",
	  dpt.dt_failcall, dpt.dt_waiting,
	  dpt.dt_active, dpt.dt_active_c, dpt.dt_active_u);
    return 1;
}

int
set_dtimeouts(rp, dlu, pkt)
REMOTE *rp;
register dl_unitdata_ind_t *dlu;
char *pkt;
{
    return set_timeouts(rp, 1, dlu, pkt);
}

int
set_atimeouts(rp)
REMOTE *rp;
{
    return set_timeouts(rp, 0, (dl_unitdata_ind_t *)0, (char *)0);
}


/*
 * Routines for version 1 access files.
 */
static char *space_sep;

static condition_t *
parse_condition_orlist(type)
int type;
{
    char *tok;
    int t;
    condition_t *c, *co, *cr;

    tok = nexttok(SPACES, &space_sep);
    if (!tok)
	return (condition_t *)0;
    c = parse_path(tok, type);
    if (cr = parse_condition_orlist(type)) {
	co = alloc_condition(COND_OR);
	co->Left = c;
	co->Right = cr;
	c = co;
    }
    return c;
}

static condition_t *
parse_condition_list(type, negate)
int type;
int negate;
{
    condition_t *c, *nc;
    char *tok;

    c = parse_condition_orlist(type);
    if (negate) {
	nc = alloc_condition(COND_NOT);
	nc->Right = c;
	return nc;
    }
    else
	return c;
}

static void
parse_timeout_v1(dt, which)
dp_timeo_t *dt;
int which;
{
    char *val = nexttok(SPACES, &space_sep);

    if (!val) {
	yyerror(access_file, "Missing timeout value");
	return;
    }

    switch (which) {
     case TO_FAILCALL:	dt->dt_failcall = parse_seconds(val);	break;
     case TO_WAITING:	dt->dt_waiting  = parse_seconds(val);	break;
     case TO_ACTIVE:	dt->dt_active   = parse_seconds(val);	break;
     case TO_ACTIVE_C:	dt->dt_active_c = parse_seconds(val);	break;
     case TO_ACTIVE_U:	dt->dt_active_u = parse_seconds(val);	break;
     default:		yyerror(access_file, "Bad timeout designation");
			break;
    }
}

static condition_t *
parse_time_list(day)
char *day;
{
    char pbuf[512];
    char *p = pbuf;
    char *hr;
    char sep = '/';
    condition_t *c, *ca;

    (void)strcpy(p, day);
    p += strlen(p);

    while (hr = nexttok(SPACES, &space_sep)) {
	*p++ = sep;
	sep = ',';
	(void)strcpy(p, hr);
	p += strlen(p);
    }
    *p = '\0';
    c = parse_time_path(pbuf);
    if (sep == '/') {
	/*
	 * This condition is guaranteed to fail.  We need to put something
	 * in here or else post processing will assume we had no conditions
	 * and all times will be assumed good.
	 */
	ca = alloc_condition(COND_AND);
	ca->Left = c;
	ca->Right = alloc_condition(COND_FALSE);
	c = ca;
    }
    return c;
}

static void
add_time_cond(cp, c)
condition_t **cp, *c;
{
    condition_t *nc;

    while (*cp && (*cp)->Type == COND_OR)
	cp = &(*cp)->Right;
    if (*cp) {
	nc = alloc_condition(COND_OR);
	nc->Left = *cp;
	nc->Right = c;
	c = nc;
    }
    *cp = c;
}

static int
parse_access_v1(ap)
access_t *ap;
{
    dp_timeo_t *ddt, *adt;
    dp_timeo_t *timeouts;
    char buff[MAXLN];
    char *word;

    ap->DTimeouts = alloc_timeout();
    ddt = &ap->DTimeouts->DefaultCondition->Timeouts;

    ap->ATimeouts = alloc_timeout();
    adt = &ap->ATimeouts->DefaultCondition->Timeouts;

    /*
     * Read access file.
     */
    while (agetln(buff, sizeof buff)) {
	if ((word = strtok_r(buff, SPACES, &space_sep)) == NULL)
	    continue;
	if (*word == '#')
	    continue;
	/*
	 * Allow the setting of Dial out or Answer call timeouts seperately.
	 * This should also be done for allowable call times, etc., but that
	 * has not yet been implemented.
	 */
	timeouts = ddt;
	if (strncmp(word, "answer:", sizeof("answer:")-1) == 0) {
	    word += sizeof("answer:")-1;
	    timeouts = adt;
	}
	else if (strncmp(word, "dial:", sizeof("dial:")-1) == 0)
	    word += sizeof("dial:")-1;

	/*
	 * Dispatch on the word to fill in the fields.
	 */
	if (EQ(word, "allowto") || EQ(word, "gooddstaddresses")) {
	    if (ap->Addrs)
		yyerror(access_file, "more than one address directive");
	    ap->Addrs = parse_condition_list(COND_ADDR, 0);
	}
	else if (EQ(word, "disallowfrom") || EQ(word, "badsrcaddresses")) {
	    if (ap->Addrs)
		yyerror(access_file, "more than one address directive");
	    ap->Addrs = parse_condition_list(COND_ADDR, 1);
	}
	else if (EQ(word, "protocols") || EQ(word, "goodprotocols")) {
	    if (ap->Protos)
		yyerror(access_file, "more than one protocol directive");
	    ap->Protos = parse_condition_list(COND_PROTO, 0);
	}
	else if (EQ(word, "badprotocols")) {
	    if (ap->Protos)
		yyerror(access_file, "more than one protocol directive");
	    ap->Protos = parse_condition_list(COND_PROTO, 1);
	}
	else if (EQ(word, "inactivity"))
	    parse_timeout_v1(timeouts, TO_ACTIVE);
	else if (EQ(word, "last_close"))
	    parse_timeout_v1(timeouts, TO_ACTIVE_C);
	else if (EQ(word, "non_tcp"))
	    parse_timeout_v1(timeouts, TO_ACTIVE_U);
	else if (EQ(word, "failedcall"))
	    parse_timeout_v1(timeouts, TO_FAILCALL);
	else if (EQ(word, "callwait"))
	    parse_timeout_v1(timeouts, TO_WAITING);
	else if (EQ(word, "weekdays") ||
		 EQ(word, "weekends") ||
		 mapname2num(word, day_ptab) >= 0)
	    add_time_cond(&ap->CallTimes, parse_time_list(word));
	else
	    yyerror(access_file, "Bad parameter \"%s\"", word);
    }
    return errorcount == 0;
}

/*
 * This function is provided in case this is being compiled on older
 * versions of Solaris (apparently this function is not provided in 2.0
 * or 2.1.  It appears to behave identical to the system function, at least
 * for the purposes of this program.
 */

char *
strtok_r(s1, s2, lasts)
char *s1;
const char *s2;
char **lasts;
{
    char *tok;
    if (!s1)
	s1 = *lasts;
    if (!s1)
	return (char *)0;
    while (*s1 && strchr(s2, *s1))
	s1++;
    if (!*s1)
	return (char *)0;
    tok = s1;
    while (*s1 && !strchr(s2, *s1))
	s1++;
    if (*s1) {
	*s1++ = '\0';
	*lasts = s1;
    }
    else
	*lasts = (char *)0;
    return tok;
}
