#include <stdio.h>
#include <string.h>
#include <netdb.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <sys/dlpi.h>

#include "ppp.h"
#include "dp.h"
#include "dpio.h"
#include "dpd.h"
#include "dpconf.h"

typedef struct pkt {
    time_t time;
    ulong ppa;
    dl_unitdata_ind_t *dlu;
    int dlu_len;
    char *pkt;
    int pkt_len;
} pkt_t;

#ifdef	__STDC__
void print_condition(condition_t *, int);
char *ppath_str(path_t *);
char *apath_str(path_t *);
char *tpath_str(path_t *);
char *path_str(path_t *);
char *time_str(long, int);
#else
void print_condition(/* condition_t *, int */);
char *ppath_str(/* path_t * */);
char *apath_str(/* path_t * */);
char *tpath_str(/* path_t * */);
char *path_str(/* path_t * */);
char *time_str(/* long, int */);
#endif
REMOTE rem;

main(argc, argv)
int argc;
char **argv;
{
    int i;
    init_pathconf();
    if (argc <= 1) {
	fprintf(stdout, "usage: %s access_file [trace_file]\n");
	exit(1);
    }
    testaccess(argv[1], argc > 2 ? argv[2] : (char *)0);
}

testaccess(path, tf)
char *path;
char *tf;
{
    pkt_t pkt;
    int a;
    dp_timeo_t to;

    (void)readaccess("test", path, &rem);
    if (rem.Access)
	print_access(rem.Access);

    if (tf) {
	open_trace(tf);
	comp_timeos(&to, rem.Access, 0, (dl_unitdata_ind_t *)0, (char *)0);
	print_timeos(&to);
	while (get_req(&pkt)) {
	    print_req(&pkt);
	    a = authorized(rem.Access, pkt.dlu, pkt.dlu_len,
			   pkt.pkt, pkt.pkt_len);
	    printf("authorized = %d\n", a);
	    comp_timeos(&to, rem.Access, 1, pkt.dlu, pkt.pkt);
	    print_timeos(&to);
	}
	close_trace();
    }
}

char *
secs_str(t)
int t;
{
    char buf[32];
    
    if (t < 0)
	switch (t) {
	 case DPT_NO_TIMEOUT:	  return "none";
	 case DPT_DEF_TIMEOUT:	  return "default";
	 case DPT_UNSPEC_TIMEOUT: return "unset";
	}

    if (t < 60)
	(void)sprintf(buf, "%d",
		      t);
    else if (t < 60*60)
	(void)sprintf(buf, "%d:%02d",
		      t / 60,
		      t % 60);
    else if (t < 24*60*60)
	(void)sprintf(buf, "%d:%02d:%02d",
		      t / (60 * 60),
		      (t % (60 * 60)) / 60,
		      t % 60);
    else
	(void)sprintf(buf, "%d:%02d:%02d:%02d",
		      t / (24 * 60 * 60),
		      (t % (24 * 60 * 60)) / (60 * 60),
		      (t % (60 * 60)) / 60,
		      t % 60);
    return buf;
}

char *
proto_str(p)
pattern_t *p;
{
    static char sbuf[256];
    char *s = sbuf;
    range_t *r;
    while (p) {
	r = p->RangeList;
	while (r) {
	    if (r->Min == r->Max)
		(void)sprintf(s, "%d", r->Min);
	    else
		(void)sprintf(s, "[%d,%d]", r->Min, r->Max);
	    s += strlen(s);
	    if (r = r->Next)
		*s++ = ',';
	}
	if (p = p->NextLevel)
	    *s++ = '/';
    } *s = '\0';
    return sbuf;
}

char *
addr_str(p)
pattern_t *p;
{
    static char sbuf[256];
    char *s = sbuf;
    range_t *r;
    int lev;
    for (lev = 0 ; p ; lev++) {
	r = p->RangeList;
	while (r) {
	    if (r->Min == r->Max)
		(void)sprintf(s, "%d", r->Min);
	    else
		(void)sprintf(s, "[%d,%d]", r->Min, r->Max);
	    s += strlen(s);
	    if (r = r->Next)
		*s++ = ',';
	}
	if (p = p->NextLevel)
	    *s++ = lev > 0 ? '.' : '/';
    } *s = '\0';
    return sbuf;
}

char *
times_str(p)
pattern_t *p;
{
    static char sbuf[256];
    char *s = sbuf;
    range_t *r;
    int lev;

    lev = 0;
    while (p) {
	r = p->RangeList;
	while (r) {
	    if (r->Min == r->Max)
		(void)sprintf(s, "%s", time_str(r->Min, lev));
	    else {
		(void)sprintf(s, "[%s,", time_str(r->Min, lev));
		s += strlen(s);
		(void)sprintf(s, "%s]",  time_str(r->Max, lev));
	    }
	    s += strlen(s);
	    if (r = r->Next)
		*s++ = ',';
	}
	lev++;
	if (p = p->NextLevel)
	    *s++ = '/';
    }
    *s = '\0';
    return sbuf;
}

char *
cond_str(c)
condition_t *c;
{
    pattern_t *p = c->Pattern;
    switch (c->Type) {
     case COND_DSTPROTO:
     case COND_SRCPROTO:
     case COND_PROTO:
	return proto_str(p);
     case COND_DSTADDR:
     case COND_SRCADDR:
     case COND_ADDR:
	return addr_str(p);
     case COND_TIME:
	return times_str(p);
     default:
	return "Unknown condition type";
    }
}


void
print_expr(c, s, lead)
condition_t *c;
char *s;
int lead;
{
    if (c->Left) {
	printf("%*s(\n", lead, "");
	print_condition(c->Left, lead + 1);
    }
    printf("%*s%s\n", lead, "", s);
    print_condition(c->Right, lead + 1);
    printf("%*s)\n", lead, "");
}

void
print_condition(c, lead)
condition_t *c;
int lead;
{
    switch (c->Type) {
     case COND_OR:
	print_expr(c, "||", lead);
	break;
     case COND_AND:
	print_expr(c, "&&", lead);
	break;
     case COND_NOT:
	print_expr(c, "!", lead);
	break;
     case COND_DSTPROTO:
	printf("%*sDst Proto %s\n", lead, "", proto_str(c->Pattern));
	break;
     case COND_SRCPROTO:
	printf("%*sSrc Proto %s\n", lead, "", proto_str(c->Pattern));
	break;
     case COND_PROTO:
	printf("%*sProto %s\n", lead, "", proto_str(c->Pattern));
	break;
     case COND_DSTADDR:
	printf("%*sDst Addr %s\n", lead, "", addr_str(c->Pattern));
	break;
     case COND_SRCADDR:
	printf("%*sSrc Addr %s\n", lead, "", addr_str(c->Pattern));
	break;
     case COND_ADDR:
	printf("%*sAddr %s\n", lead, "", addr_str(c->Pattern));
	break;
     case COND_TIME:
	printf("%*s%s\n", lead, "", times_str(c->Pattern));
	break;
     case COND_TRUE:
	printf("%*strue\n", lead, "");
	break;
     case COND_FALSE:
	printf("%*sfalse\n", lead, "");
	break;
     default:
	printf("%*sUnknown Condition %d\n", lead, "", c->Type); 
	break;
    }
}

void
print_conditions(c, cs)
condition_t *c;
char *cs;
{
    if (!c)
	return;
    printf("%s:\n", cs);
    print_condition(c, 2); 
}

void
print_timeouts(to, ts)
timeout_t *to;
char *ts;
{
    condition_list_t *cl;
    condition_t *c;
    dp_timeo_t *dt;

    printf("%s Timeouts:\n", ts);
    if (!to) {
	printf("      none\n");
	return;
    }
    for (cl = to->ConditionList ; cl ; cl = cl->Next) {
	if (cl->Conditions)
	    print_condition(cl->Conditions, 6);
	else
	    printf("  No Conditions\n");
	dt = &cl->Timeouts;
	printf("  Timeouts:\n");
	if (dt->dt_failcall != DPT_UNSPEC_TIMEOUT)
	    printf("    failedcall %8s\n", secs_str(dt->dt_failcall));
	if (dt->dt_waiting != DPT_UNSPEC_TIMEOUT)
	    printf("    callwait   %8s\n", secs_str(dt->dt_waiting));
	if (dt->dt_active != DPT_UNSPEC_TIMEOUT)
	    printf("    inactivity %8s\n", secs_str(dt->dt_active));
	if (dt->dt_active_c != DPT_UNSPEC_TIMEOUT)
	    printf("    last_close %8s\n", secs_str(dt->dt_active_c));
	if (dt->dt_active_u != DPT_UNSPEC_TIMEOUT)
	    printf("    non_stream %8s\n", secs_str(dt->dt_active_u));
    }

}

print_timeos(dt)
dp_timeo_t *dt;
{
    if (!dt) {
	printf("none\n");
	return;
    }
    printf("failcall %-5s ",  secs_str(dt->dt_failcall));
    printf("waiting  %-5s ",  secs_str(dt->dt_waiting));
    printf("active   %-5s ",  secs_str(dt->dt_active));
    printf("active_c %-5s ",  secs_str(dt->dt_active_c));
    printf("active_u %-5s\n", secs_str(dt->dt_active_u));
}


print_access(ap)
access_t *ap;
{
    printf("----------\n");
    printf("File Name: %s\n", ap->Path);
    print_conditions(ap->Protos, "Allowed Protocols");
    print_conditions(ap->CallTimes, "Allowed Call Times");
    print_conditions(ap->ConnectTimes, "Continuous Call Periods");
    print_conditions(ap->Addrs, "Allowed Addresses");
    print_timeouts(ap->DTimeouts, "Dial");
    print_timeouts(ap->ATimeouts, "Answer");
    print_timeouts(ap->Timeouts, "Both");
}

int linenum;
int errorcount;

#include <varargs.h>

/* VARARGS2 */
void
yyerror(va_alist)
va_dcl
{
    va_list ap;
    char *path,
	 *fmt,
	 buff[256];

    va_start(ap);
    path = va_arg(ap, char *);
    fmt = va_arg(ap, char *);
    (void)sprintf(buff, "Error near line %d of\n\t\"%s\":  %s\n",
	linenum, path, fmt);
    vfprintf(stdout, buff, ap);
    va_end(ap);
    errorcount++;
}

/* VARARGS0 */
void
d_log(va_alist)
    va_dcl
{
    int		level;
    char	*where;
    char	*format;
    va_list	ap;

    va_start(ap);
    level = va_arg(ap, int);
    where = va_arg(ap, char*);
    format = va_arg(ap, char*);
    fprintf(stdout, "%s: ", where);
    vfprintf(stdout, format, ap);
    fprintf(stdout, "\n");
    va_end(ap);
}

int pkt_buf[1024+PPP_MTU/sizeof(int)];

FILE *rf;

open_trace(f)
char *f;
{
    char *tf;

    tf = expand_dir_file("$DPLOG_DIR", f);
    rf = fopen(tf, "r");

    if (rf == NULL) {
	fprintf(stdout, "can't open trace file \"%s\"\n", tf);
	exit(1);
    }
}

close_trace()
{
    (void)fclose(rf);
}

get_req(pkt)
pkt_t *pkt;
{
    int n;
    pkt_trace_t pt;
    if (!fread((char *)&pt, sizeof(pt), 1, rf))
	return 0;
    if (pt.dlu_len + pt.pkt_len > sizeof(pkt_buf)) {
	fprintf(stdout, "corrupted trace file\n");
	return 0;
    }
    pkt->time = pt.time;
    pkt->ppa  = pt.ppa;
    pkt->dlu  = (dl_unitdata_ind_t *)pkt_buf;
    pkt->dlu_len = pt.dlu_len;
    pkt->pkt  = ((char *)pkt_buf) + pt.dlu_len;
    pkt->pkt_len = pt.pkt_len;
    if (!fread((char *)pkt->dlu, pt.dlu_len, 1, rf))
	return 0;
    return fread((char *)pkt->pkt, pt.pkt_len, 1, rf);
}

print_req(pkt)
pkt_t *pkt;
{
    ulong dlsap;
    phdr_t phdr;
    phdr_t *ph;
    printf("\n");
    dlsap = *(ulong *)(((char *)pkt->dlu) + pkt->dlu->dl_dest_addr_offset);
    switch (DP_DLSAP_PROTO(dlsap) & ~DPP_DIAL_REQ) {

     default:
	/*
	 * We should not get an unsupported protocol....
	 */
	dlsap2ph(&phdr, dlsap);
	printf("Unknown protocol %02x:%02x:%04x\n",
	      phdr.ph_address,
	      phdr.ph_control,
	      ntohs(phdr.ph_protocol));

     case PPP_IP:
	print_ip((struct ip *)pkt->pkt);
	break;
    }
}

print_ip(ip)
struct ip *ip;
{
    struct udphdr *uh;
    char *proto, pbuf[32];
    switch (ip->ip_p) {
     case IPPROTO_TCP:
	proto = "TCP";
	break;
     case IPPROTO_UDP:
	proto = "UDP";
	break;
     case IPPROTO_ICMP:
	proto = "ICMP";
	break;
     default:
	(void)sprintf(pbuf, "%d", ip->ip_p);
	proto = pbuf;
	break;
    }
    printf("%-4s ", proto);
    switch (ip->ip_p) {
     case IPPROTO_TCP:
	{
	    struct tcphdr *th = (struct tcphdr *)&((char *)ip)[ip->ip_hl << 2];
	    printf("%s.%d -> ", inet_ntoa(ip->ip_src), th->th_sport);
	    printf("%s.%d\n",   inet_ntoa(ip->ip_dst), th->th_dport);
	}
	break;
     case IPPROTO_UDP:
	{
	    struct udphdr *uh = (struct udphdr *)&((char *)ip)[ip->ip_hl << 2];
	    printf("%s.%d -> ", inet_ntoa(ip->ip_src), uh->uh_sport);
	    printf("%s.%d\n",   inet_ntoa(ip->ip_dst), uh->uh_dport);
	}
	break;
     default:
	printf("%s -> ", inet_ntoa(ip->ip_src));
	printf("%s\n",   inet_ntoa(ip->ip_dst));
	break;
    }
}

print_sin(sin)
struct sockaddr_in *sin;
{
    printf("dst: family: %2d  ", sin->sin_family);
    printf("addr: %s\n", inet_ntoa(sin->sin_addr));
}
