/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/gn.c
    Version 1.0
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include "gn.h"

static char *www_preparse();


main( argc, argv)
int	argc;
char	*argv[];

{
	char	selector[PATHLEN],
		*selp = selector;
	int	itemtype;
	Item	thisitem;

	gn_init( argc, argv);

	get_selector( selector);
	strcpy( thisitem.selector, selector);

	/* If the first  letters is in the following list it should
	   be an HTTP request					   */

	switch ( *selector) {
		case 'G':
		case 'H':
		case 'S':
		case 'P':
		case 'T':
			if ( (selp = www_preparse( selector)) == NULL)
				exit(0);
	}

	parse_selector( &thisitem, selp, FULL);

	if ( streq( thisitem.gtype, "0v")) {
		send_text_line( VERSION);
		send_end();
		return (0);
	}

	if ( chkaccess( &thisitem, LOG))
		itemtype = chkcache( &thisitem);
	else
		itemtype = DENIED;

	switch (itemtype) {
		case ROOT:
			sendcache( &thisitem);
			break;
		case TEXT:
			sendtext( &thisitem);
			break;
		case RANGE:
			sendrange( &thisitem);
			break;
		case DIR:
			sendcache( &thisitem);
			break;
		case BIN:
			sendbin( &thisitem);
			break;
		case EXEC:
			sendexec( &thisitem);
			break;
		case SEARCH:
			sendsearch( &thisitem);
			break;
		case COMMENT:
			exit( 0);
			break;
		case DENIED:
			senderr( "Sorry, access denied", &thisitem);
			break;
		case UNKNOWN:
			senderr( "Unknown selector type", &thisitem);
			break;
		default:
			senderr( "Unknown server error", &thisitem);
	}
	return (0);
}


static char *
www_preparse( sel)
char	*sel;
{
	register char	*cp;
	char		errmsg[BUFSIZE];
	int isgif = FALSE;

	iswww = TRUE;
	if ( strncmp( sel, "GET", 3) == 0) {
		sel += 3;
		while (  *sel && isspace( *sel))
			sel++;
		cp = sel;
		while (  *cp && !isspace( *cp))
			cp++;
		*cp = '\0';

		if ( streq ( sel + strlen(sel) - 4, ".gif" ))
			isgif = TRUE;

		www_unescape( sel);

		if ( *sel == '/' )
			sel++;
		if ( (cp = strrchr( sel, '?')) != NULL)
			*cp = '\t';

		/* Disgusting hack! */
		if ( isgif ) {
			*sel = ( (*sel == '0') ? '9' : *sel );
			*sel = ( (*sel == '1') ? '9' : *sel );
		}

	}
	else {
		if ( (cp = strchr( sel, ' ')) != NULL)
			*cp = '\0';
		strcpy( errmsg, sel);
		strcat( errmsg, ": Method not implemented");
		www_err( "501", errmsg);
		sel = NULL;
	}
	return sel;
}
