/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/send.c
    Version 1.0
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include "gn.h"
#include "search.h"

static int	ishtml();

/*
 * sendtext( ip)  Send the text file with Item pointer ip.
 */

void
sendtext( ip)
Item	*ip;

{
	FILE	*fp;
	char	linebuf[BUFSIZE],
		commandbuf[BUFSIZE];


	if ( ip->compressed) {
		sprintf( commandbuf, "%s %s", DECOMPRESS, ip->filepath);
		if ( (fp = popen( commandbuf, "r")) == (FILE *) NULL ) {
			senderr2( "Can't decompress", ip->filepath);
			exit( 2);
		}
	}
	else {

		/* If an item has type "0h" (zero h) that means that it has
        	   two versions: a plaintext version named "whatever/file" and 
	           an html version nameed "whatever/file.html".  If this is an
	           http transaction we use the latter.
	         */
		if ( iswww && streq( ip->gtype, "0h"))
			strcat( ip->filepath, ".html");

		if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
			senderr2( "Can't open file", ip->filepath);
			exit( 2);
		}
	}

	if ( iswww && !ishtml( ip->filepath))
		send_text_line( "<pre>");

	while ( fgets( linebuf, BUFSIZE-2, fp))
		send_text_line( linebuf);

	if ( iswww && !ishtml( ip->filepath))
		send_text_line( "</pre>");

	if ( !iswww )
		send_end();

	writelog( "Sent text", ip->selector);

	if ( ip->compressed)
		pclose( fp);
	else
		fclose( fp);
}


/*
 * ishtml( path)  True if string path ends in .html
 */

static int
ishtml( path)
char	*path;
{
	register char	*cp;

	if ( (cp = strrchr( path, '.')) == NULL)
		return FALSE;
	else
		return (streq( cp, ".html"));
}



/*
 * sendrange( ip)  Send a range from a textfile
 */

void
sendrange( ip)
Item	*ip;

{
	FILE	*fp;

	char	linebuf[BUFSIZE];

	if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open file", ip->filepath);
		exit( 2);
	}

	if ( iswww && !ishtml( ip->filepath))
		send_text_line( "<pre>");

	fseek( fp, ip->range_start, 0);
	while ( fgets( linebuf, BUFSIZE, fp) && (ftell( fp) <= ip->range_end)) {
		chop( linebuf);
		if ( streq( linebuf, "."))
			strcat( linebuf, ".");  
			/* Double a '.' if on line by itself */
		send_text_line( linebuf);
	}	

	if ( iswww && !ishtml( ip->filepath))
		send_text_line( "</pre>");
	if ( !iswww )
		send_end();
	writelog( "Sent range", ip->selector);
}




/*
 * sendbin( ip)  Send a binary file.
 */

void
sendbin(  ip)
Item	*ip;

{
	FILE	*fp;
	char	commandbuf[PATHLEN];
	int	c;


	if ( ip->compressed) {
		sprintf( commandbuf, "%s %s", DECOMPRESS, ip->filepath);
		if ( (fp = popen( commandbuf, "r")) == (FILE *) NULL ) {
			senderr2( "Can't decompress", ip->filepath);
			exit( 2);
		}
	}
	else
		if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
			senderr2( "Can't open file", ip->filepath);
			exit( 2);
		}
	
	while ( (c = fgetc( fp )) != EOF)
		fputc( c, stdout);
	writelog( "Sent binary", ip->selector);

	if ( ip->compressed)
		pclose( fp);
	else
		fclose( fp);
}

/*
 * sendcache( ip)  Send the contents of the cachefile
 * in the directory pointed to by selector.  ip->filepath is the complete
 * pathname of the directory with cfname appended.
 */

void
sendcache(  ip)
Item	*ip;

{
	FILE	*fp;

	char	cline[BUFSIZE];

	if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open cache file (sendcache): ", 
			ip->filepath);
		exit( 2);
	}
	www_prolog( ip, "Menu");
	while ( get_cache_line( cline, fp)) {
		if ( iswww)
			www_cache_line( cline);
		else
			send_cache_line( cline);
	}
	cache_epilog(iswww);

	writelog( "Sent cache", ip->selector);
}


void
cache_epilog( iswww)
int	iswww;
{
	if ( iswww)
		send_text_line( "</ul>" );
	else
		send_end();
}


void
www_prolog(ip, str)
Item	*ip;
char	*str;
{

	char	*cp,
		buf[BUFSIZE];

	if ( !iswww)
		return;
	if ( ((*ip->cachepath) == '\0') && ((*ip->name) == '\0') )
		/* It's the root dir and there's no ip->name */
		sprintf(buf,"<ul>");
	else {
		cp = ( (*(ip->name)) ? ip->name : str );
		sprintf(buf,"<title> %s </title><h2> %s</h2><ul>", cp, cp);
	}
	send_text_line( buf);
}	


/*
 * sendexec( ip)  Open pipe from "ip->filepath" command
 * and send output.
 */

void
sendexec(  ip)
Item	*ip;
{
	FILE	*fp;
	register int	c;
	char	comargs[PATHLEN],
		linebuf[PATHLEN];


	strcpy(comargs, ip->args);
	if ( iswww)
		strcat( comargs, " \"www\"");
	if ( (fp = safer_popen( ip->filepath, comargs, "r"))
				 == (FILE *) NULL ) {
		senderr2( "Can't exec", ip->filepath);
		exit( 2);
	}

	switch (*(ip->gtype + 4)) {  /* the char after "exec" in gtype */
		case '1':
			www_prolog( ip, "Menu");
			while ( get_cache_line( linebuf, fp)) {
				if ( iswww)
					www_cache_line( linebuf);
				else
					send_cache_line( linebuf);
			}
			cache_epilog(iswww);
			break;
		case '5':
		case '9':
		case 's':
		case 'I':
			while ( (c = fgetc( fp )) != EOF)
				fputc( c, stdout);
			pclose( fp);
			writelog( "Exec binary", ip->selector);
			return;
		case '0':
		case '4':
		case ':':
		default:
			if ( iswww )
				send_text_line( "<pre>");
			while ( fgets( linebuf, BUFSIZE, fp))
				send_text_line( linebuf);
			if ( iswww )
				send_text_line( "</pre>");
			else
				send_end();
			break;
	}

	pclose( fp);
	writelog( "Exec", ip->selector);
}
