/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/send.c
    Version 1.1
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include "gn.h"
#include "search.h"


void
sendhead( ip)
Item	*ip;

{
	http_prolog( ip);
}


/*
 * sendtext( ip)  Send the text file with Item pointer ip.
 */

void
sendtext( ip)
Item	*ip;

{
	FILE	*fp;
	char	linebuf[BUFSIZE],
		commandbuf[BUFSIZE];
	int	send_pre = FALSE;


	if ( ip->compressed) {
		sprintf( commandbuf, "%s %s", DECOMPRESS, ip->filepath);
		if ( (fp = popen( commandbuf, "r")) == (FILE *) NULL ) {
			senderr2( "Can't decompress", ip->filepath);
			exit( 2);
		}
	}
	else {

		/* If an item has type "0h" (zero h) that means that it has
        	   two versions: a plaintext version named "whatever/file" and 
	           an html version nameed "whatever/file.html".  If this is an
	           http transaction we use the latter.
	         */
		if ( http && streq( ip->gtype, "0h")) {
			strcpy( ip->content_type, "text/html");
			strcpy( ip->suffix, "html");
			strcat( ip->filepath, ".html");
		}

		if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
			senderr2( "Can't open file", ip->filepath);
			exit( 2);
		}
	}

	http_prolog( ip);
	if ( http == HTTP0_9 && streq( ip->content_type, "text/plain") &&
				!streq( ip->suffix, "txt") )
		send_pre = TRUE;

	if ( send_pre)
		send_text_line( "<BODY><PRE>");

	while ( fgets( linebuf, BUFSIZE-2, fp))
		send_text_line( linebuf);

	if ( send_pre)
		send_text_line( "</PRE></BODY>");

	if ( !http )
		send_end();

	writelog( "Sent text", ip->selector);

	if ( ip->compressed)
		pclose( fp);
	else
		fclose( fp);
}



/*
 * sendrange( ip)  Send a range from a textfile
 */

void
sendrange( ip)
Item	*ip;

{
	FILE	*fp;
	int	send_pre = FALSE;

	char	linebuf[BUFSIZE];

	if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open file", ip->filepath);
		exit( 2);
	}

	http_prolog( ip);

	if ( http == HTTP0_9 && streq( ip->content_type, "text/plain") &&
				!streq( ip->suffix, "txt") )
		send_pre = TRUE;

	if ( send_pre)
		send_text_line( "<BODY><PRE>");

	fseek( fp, ip->range_start, 0);
	while ( fgets( linebuf, BUFSIZE, fp)
				&& (ftell( fp) <= ip->range_end))
		send_text_line( linebuf);

	if ( send_pre )
		send_text_line( "</PRE></BODY>");
	if ( !http )
		send_end();
	writelog( "Sent range", ip->selector);
}




/*
 * sendbin( ip)  Send a binary file.
 */

void
sendbin(  ip)
Item	*ip;

{
	FILE	*fp;
	char	commandbuf[PATHLEN];
	int	c;


	if ( ip->compressed) {
		sprintf( commandbuf, "%s %s", DECOMPRESS, ip->filepath);
		if ( (fp = popen( commandbuf, "r")) == (FILE *) NULL ) {
			senderr2( "Can't decompress", ip->filepath);
			exit( 2);
		}
	}
	else
		if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
			senderr2( "Can't open file", ip->filepath);
			exit( 2);
		}
	
	http_prolog( ip);
	while ( (c = fgetc( fp )) != EOF)
		fputc( c, stdout);
	writelog( "Sent binary", ip->selector);

	if ( ip->compressed)
		pclose( fp);
	else
		fclose( fp);
}

/*
 * sendcache( ip)  Send the contents of the cachefile
 * in the directory pointed to by selector.  ip->filepath is the complete
 * pathname of the directory with cfname appended.
 */

void
sendcache(  ip)
Item	*ip;

{
	FILE	*fp;

	char	*cline,
		*dummy;

	if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open cache file (sendcache): ", 
			ip->filepath);
		exit( 2);
	}
	cache_prolog( ip, "Menu", "");
	while ( read_cache( &cline, &dummy, fp)) {
		if ( http)
			www_cache_line( cline);
		else
			send_cache_line( cline);
	}
	cache_epilog(http);

	writelog( "Sent cache", ip->selector);
}


void
cache_epilog( http)
int	http;
{
	if ( http)
		send_text_line( "</ul></body>" );
	else
		send_end();
}


void
cache_prolog(ip, str, option)
Item	*ip;
char	*str,
	*option;
{

	char	*cp,
		buf[BUFSIZE];

	if ( !http )
		return;
	http_prolog( ip);

	if ( ((*ip->cachepath) == '\0') && ((*ip->name) == '\0') ) {
		/* It's the root dir and there's no ip->name */
		sprintf(buf,
			"<head><title>%s%s root menu</title></head>",
				option, GN_HOSTNAME);
		strcat(buf,"<body><ul>");
		send_text_line( buf);
	}
	else {
		cp = ( (*(ip->name)) ? ip->name : str );
		sprintf(buf,"<head>%s<title> %s </title></head>", option, cp);
		send_text_line( buf);
		sprintf(buf,"<body><h2> %s</h2><ul>", cp);
		send_text_line( buf);
	}
}	


/*
 * sendexec( ip)  Open pipe from "ip->filepath" command
 * and send output.
 */

void
sendexec(  ip)
Item	*ip;
{
	FILE	*fp;
	register int	c;
	int	send_pre = FALSE;
	char	comargs[PATHLEN],
		buf[MAXLEN],
		*linebuf,
		*dummy;


	strcpy(comargs, ip->args);
	if ( http)
		strcat( comargs, " \"www\"");
	if ( (fp = safer_popen( ip->filepath, comargs, "r"))
				 == (FILE *) NULL ) {
		senderr2( "Can't exec", ip->filepath);
		exit( 2);
	}

	switch (*(ip->gtype + 4)) {  /* the char after "exec" in gtype */
		case '1':
			cache_prolog( ip, "Menu", "");
			while ( read_cache( &linebuf, &dummy, fp)) {
				if ( http)
					www_cache_line( linebuf);
				else
					send_cache_line( linebuf);
			}
			cache_epilog(http);
			break;
		case '5':
		case '9':
		case 's':
		case 'I':
			http_prolog( ip);
			while ( (c = fgetc( fp )) != EOF)
				fputc( c, stdout);
			pclose( fp);
			writelog( "Exec binary", ip->selector);
			return;
		case '0':
		case '4':
		case ':':
		default:
			http_prolog( ip);
			if ( http == HTTP0_9 && 
				streq( ip->content_type, "text/plain") &&
					!streq( ip->suffix, "txt") )
				send_pre = TRUE;
	
			if ( send_pre)
				send_text_line( "<BODY><PRE>");

			while ( fgets( buf, BUFSIZE, fp))
				send_text_line( buf);
			if ( send_pre )
				send_text_line( "</PRE></BODY>");
			else if ( !http)
				send_end();
			break;
	}

	pclose( fp);
	writelog( "Exec", ip->selector);
}

void
http_prolog( ip)
Item	*ip;
{

	char	buf[BUFSIZE];

	if ( http & HTTP1_0) {
		sprintf( buf, "%s 200 Request fulfilled", HTTPVERSION);
		send_text_line( buf);
		sprintf( buf, "Server: %s", VERSION);
		send_text_line( buf);
		sprintf( buf, "MIME-version: %s", "1.0");
		send_text_line( buf);
		if ( *ip->mod_date) {
			sprintf( buf, "Last-Modified: %s", ip->mod_date);
			send_text_line( buf);
		}
		if ( *ip->length) {
			sprintf( buf, "Content-length: %s", ip->length);
			send_text_line( buf);
		}
		sprintf( buf, "Content-type: %s", ip->content_type);
		send_text_line( buf);
		send_text_line( "");
	}
}
